"""Base module to interact with Revit ribbon."""
from collections import OrderedDict

#pylint: disable=W0703,C0302,C0103
from pyrevit import HOST_APP, EXEC_PARAMS, PyRevitException
from pyrevit.compat import safe_strtype
from pyrevit import coreutils
from pyrevit.coreutils.logger import get_logger
from pyrevit.coreutils import envvars
from pyrevit.framework import System, Uri, Windows
from pyrevit.framework import IO
from pyrevit.framework import Imaging
from pyrevit.framework import BindingFlags
from pyrevit.framework import Media, Convert
from pyrevit.api import UI, AdWindows, AdInternal
from pyrevit.runtime import types
from pyrevit.revit import ui


mlogger = get_logger(__name__)


PYREVIT_TAB_IDENTIFIER = 'pyrevit_tab'

ICON_SMALL = 16
ICON_MEDIUM = 24
ICON_LARGE = 32

DEFAULT_DPI = 96

DEFAULT_TOOLTIP_IMAGE_FORMAT = '.png'
DEFAULT_TOOLTIP_VIDEO_FORMAT = '.swf'
if HOST_APP.is_newer_than(2019, or_equal=True):
    DEFAULT_TOOLTIP_VIDEO_FORMAT = '.mp4'


def argb_to_brush(argb_color):
    # argb_color is formatted as #AARRGGBB
    a = r = g = b = "FF"
    try:
        b = argb_color[-2:]
        g = argb_color[-4:-2]
        r = argb_color[-6:-4]
        if len(argb_color) > 7:
            a = argb_color[-8:-6]
        return Media.SolidColorBrush(Media.Color.FromArgb(
                Convert.ToInt32("0x" + a, 16),
                Convert.ToInt32("0x" + r, 16),
                Convert.ToInt32("0x" + g, 16),
                Convert.ToInt32("0x" + b, 16)
                )
            )
    except Exception as color_ex:
        mlogger.error("Bad color format %s | %s", argb_color, color_ex)


def load_bitmapimage(image_file):
    """Load given png file.

    Args:
        image_file (str): image file path

    Returns:
        (Imaging.BitmapImage): bitmap image object
    """
    bitmap = Imaging.BitmapImage()
    bitmap.BeginInit()
    bitmap.UriSource = Uri(image_file)
    bitmap.CacheOption = Imaging.BitmapCacheOption.OnLoad
    bitmap.CreateOptions = Imaging.BitmapCreateOptions.IgnoreImageCache
    bitmap.EndInit()
    bitmap.Freeze()
    return bitmap


# Helper classes and functions -------------------------------------------------
class PyRevitUIError(PyRevitException):
    """Common base class for all pyRevit ui-related exceptions."""
    pass


class ButtonIcons(object):
    """pyRevit ui element icon.

    Upon init, this type reads the given image file into an io stream and
    releases the os lock on the file.

    Args:
        image_file (str): image file path to be used as icon

    Attributes:
        icon_file_path (str): icon image file path
        filestream (IO.FileStream): io stream containing image binary data
    """
    def __init__(self, image_file):
        self.icon_file_path = image_file
        self.check_icon_size()
        self.filestream = IO.FileStream(image_file,
                                        IO.FileMode.Open,
                                        IO.FileAccess.Read)

    @staticmethod
    def recolour(image_data, size, stride, color):
        # FIXME: needs doc, and argument types
        # ButtonIcons.recolour(image_data, image_size, stride, 0x8e44ad)
        step = stride / size
        for i in range(0, stride, step):
            for j in range(0, stride, step):
                idx = (i * size) + j
                # R = image_data[idx+2]
                # G = image_data[idx+1]
                # B = image_data[idx]
                # luminance = (0.299*R + 0.587*G + 0.114*B)
                image_data[idx] = color >> 0 & 0xff       # blue
                image_data[idx+1] = color >> 8 & 0xff     # green
                image_data[idx+2] = color >> 16 & 0xff    # red

    def check_icon_size(self):
        """Verify icon size is within acceptable range."""
        image = System.Drawing.Image.FromFile(self.icon_file_path)
        image_size = max(image.Width, image.Height)
        if image_size > 96:
            mlogger.warning('Icon file is too large. Large icons adversely '
                            'affect the load time since they need to be '
                            'processed and adjusted for screen scaling. '
                            'Keep icons at max 96x96 pixels: %s',
                            self.icon_file_path)

    def create_bitmap(self, icon_size):
        """Resamples image and creates bitmap for the given size.

        Icons are assumed to be square.

        Args:
            icon_size (int): icon size (width or height)

        Returns:
            (Imaging.BitmapSource): object containing image data at given size
        """
        mlogger.debug('Creating %sx%s bitmap from: %s',
                      icon_size, icon_size, self.icon_file_path)
        adjusted_icon_size = icon_size * 2
        adjusted_dpi = DEFAULT_DPI * 2
        screen_scaling = HOST_APP.proc_screen_scalefactor

        self.filestream.Seek(0, IO.SeekOrigin.Begin)
        base_image = Imaging.BitmapImage()
        base_image.BeginInit()
        base_image.StreamSource = self.filestream
        base_image.DecodePixelHeight = int(adjusted_icon_size * screen_scaling)
        base_image.EndInit()
        self.filestream.Seek(0, IO.SeekOrigin.Begin)

        image_size = base_image.PixelWidth
        image_format = base_image.Format
        image_byte_per_pixel = int(base_image.Format.BitsPerPixel / 8)
        palette = base_image.Palette

        stride = int(image_size * image_byte_per_pixel)
        array_size = stride * image_size
        image_data = System.Array.CreateInstance(System.Byte, array_size)
        base_image.CopyPixels(image_data, stride, 0)

        scaled_size = int(adjusted_icon_size * screen_scaling)
        scaled_dpi = int(adjusted_dpi * screen_scaling)
        bitmap_source = \
            Imaging.BitmapSource.Create(scaled_size, scaled_size,
                                        scaled_dpi, scaled_dpi,
                                        image_format,
                                        palette,
                                        image_data,
                                        stride)
        return bitmap_source

    @property
    def small_bitmap(self):
        """Resamples image and creates bitmap for size :obj:`ICON_SMALL`.

        Returns:
            (Imaging.BitmapSource): object containing image data at given size
        """
        return self.create_bitmap(ICON_SMALL)

    @property
    def medium_bitmap(self):
        """Resamples image and creates bitmap for size :obj:`ICON_MEDIUM`.

        Returns:
            (Imaging.BitmapSource): object containing image data at given size
        """
        return self.create_bitmap(ICON_MEDIUM)

    @property
    def large_bitmap(self):
        """Resamples image and creates bitmap for size :obj:`ICON_LARGE`.

        Returns:
            (Imaging.BitmapSource): object containing image data at given size
        """
        return self.create_bitmap(ICON_LARGE)


# Superclass to all ui item classes --------------------------------------------
class GenericPyRevitUIContainer(object):
    """Common type for all pyRevit ui containers.

    Attributes:
        name (str): container name
        itemdata_mode (bool): if container is wrapping UI.*ItemData
    """
    def __init__(self):
        self.name = ''
        self._rvtapi_object = None
        self._sub_pyrvt_components = OrderedDict()
        self.itemdata_mode = False
        self._dirty = False
        self._visible = None
        self._enabled = None

    def __iter__(self):
        return iter(self._sub_pyrvt_components.values())

    def __repr__(self):
        return 'Name: {} RevitAPIObject: {}'.format(self.name,
                                                    self._rvtapi_object)

    def _get_component(self, cmp_name):
        try:
            return self._sub_pyrvt_components[cmp_name]
        except KeyError:
            raise PyRevitUIError('Can not retrieve item {} from {}'
                                 .format(cmp_name, self))

    def _add_component(self, new_component):
        self._sub_pyrvt_components[new_component.name] = new_component

    def _remove_component(self, expired_cmp_name):
        try:
            self._sub_pyrvt_components.pop(expired_cmp_name)
        except KeyError:
            raise PyRevitUIError('Can not remove item {} from {}'
                                 .format(expired_cmp_name, self))

    @property
    def visible(self):
        """Is container visible."""
        if hasattr(self._rvtapi_object, 'Visible'):
            return self._rvtapi_object.Visible
        elif hasattr(self._rvtapi_object, 'IsVisible'):
            return self._rvtapi_object.IsVisible
        else:
            return self._visible

    @visible.setter
    def visible(self, value):
        if hasattr(self._rvtapi_object, 'Visible'):
            self._rvtapi_object.Visible = value
        elif hasattr(self._rvtapi_object, 'IsVisible'):
            self._rvtapi_object.IsVisible = value
        else:
            self._visible = value

    @property
    def enabled(self):
        """Is container enabled."""
        if hasattr(self._rvtapi_object, 'Enabled'):
            return self._rvtapi_object.Enabled
        elif hasattr(self._rvtapi_object, 'IsEnabled'):
            return self._rvtapi_object.IsEnabled
        else:
            return self._enabled

    @enabled.setter
    def enabled(self, value):
        if hasattr(self._rvtapi_object, 'Enabled'):
            self._rvtapi_object.Enabled = value
        elif hasattr(self._rvtapi_object, 'IsEnabled'):
            self._rvtapi_object.IsEnabled = value
        else:
            self._enabled = value

    def process_deferred(self):
        try:
            if self._visible is not None:
                self.visible = self._visible
        except Exception as visible_err:
            raise PyRevitUIError('Error setting .visible {} | {} '
                                 .format(self, visible_err))

        try:
            if self._enabled is not None:
                self.enabled = self._enabled
        except Exception as enable_err:
            raise PyRevitUIError('Error setting .enabled {} | {} '
                                 .format(self, enable_err))

    def get_rvtapi_object(self):
        """Return underlying Revit API object for this container."""
        # FIXME: return type
        return self._rvtapi_object

    def set_rvtapi_object(self, rvtapi_obj):
        """Set underlying Revit API object for this container.

        Args:
            rvtapi_obj (obj): Revit API container object
        """
        # FIXME: rvtapi_obj type
        self._rvtapi_object = rvtapi_obj
        self.itemdata_mode = False
        self._dirty = True

    def get_adwindows_object(self):
        """Return underlying AdWindows API object for this container."""
        # FIXME: return type
        rvtapi_obj = self._rvtapi_object
        getRibbonItemMethod = \
            rvtapi_obj.GetType().GetMethod(
                'getRibbonItem',
                BindingFlags.NonPublic | BindingFlags.Instance
                )
        if getRibbonItemMethod:
            return getRibbonItemMethod.Invoke(rvtapi_obj, None)

    def get_flagged_children(self, state=True):
        """Get all children with their flag equal to given state.

        Flagging is a mechanism to mark certain containers. There are various
        reasons that container flagging might be used e.g. marking updated
        containers or the ones in need of an update or removal.

        Args:
            state (bool): flag state to filter children

        Returns:
            (list[*]): list of filtered child objects
        """
        # FIXME: return type
        flagged_cmps = []
        for component in self:
            flagged_cmps.extend(component.get_flagged_children(state))
            if component.is_dirty() == state:
                flagged_cmps.append(component)
        return flagged_cmps

    def keys(self):
        # FIXME: what does this do?
        list(self._sub_pyrvt_components.keys())

    def values(self):
        # FIXME: what does this do?
        list(self._sub_pyrvt_components.values())

    @staticmethod
    def is_native():
        """Is this container generated by pyRevit or is native."""
        return False

    def is_dirty(self):
        """Is dirty flag set."""
        if self._dirty:
            return self._dirty
        else:
            # check if there is any dirty child
            for component in self:
                if component.is_dirty():
                    return True
            return False

    def set_dirty_flag(self, state=True):
        """Set dirty flag to given state.

        See .get_flagged_children()

        Args:
            state (bool): state to set flag
        """
        self._dirty = state

    def contains(self, pyrvt_cmp_name):
        """Check if container contains a component with given name.

        Args:
            pyrvt_cmp_name (str): target component name
        """
        return pyrvt_cmp_name in self._sub_pyrvt_components.keys()

    def find_child(self, child_name):
        """Find a component with given name in children.

        Args:
            child_name (str): target component name

        Returns:
            (Any): component object if found, otherwise None
        """
        for sub_cmp in self._sub_pyrvt_components.values():
            if child_name == sub_cmp.name:
                return sub_cmp
            elif hasattr(sub_cmp, 'ui_title') \
                    and child_name == sub_cmp.ui_title:
                return sub_cmp

            component = sub_cmp.find_child(child_name)
            if component:
                return component

        return None

    def activate(self):
        """Activate this container in ui."""
        try:
            self.enabled = True
            self.visible = True
            self._dirty = True
        except Exception:
            raise PyRevitUIError('Can not activate: {}'.format(self))

    def deactivate(self):
        """Deactivate this container in ui."""
        try:
            self.enabled = False
            self.visible = False
            self._dirty = True
        except Exception:
            raise PyRevitUIError('Can not deactivate: {}'.format(self))

    def get_updated_items(self):
        # FIXME: reduntant, this is a use case and should be on uimaker side?
        return self.get_flagged_children()

    def get_unchanged_items(self):
        # FIXME: reduntant, this is a use case and should be on uimaker side?
        return self.get_flagged_children(state=False)

    def reorder_before(self, item_name, ritem_name):
        """Reorder and place item_name before ritem_name.

        Args:
            item_name (str): name of component to be moved
            ritem_name (str): name of component that should be on the right
        """
        apiobj = self.get_rvtapi_object()
        litem_idx = ritem_idx = None
        if hasattr(apiobj, 'Panels'):
            for item in apiobj.Panels:
                if item.Source.AutomationName == item_name:
                    litem_idx = apiobj.Panels.IndexOf(item)
                elif item.Source.AutomationName == ritem_name:
                    ritem_idx = apiobj.Panels.IndexOf(item)
            if litem_idx and ritem_idx:
                if litem_idx < ritem_idx:
                    apiobj.Panels.Move(litem_idx, ritem_idx - 1)
                elif litem_idx > ritem_idx:
                    apiobj.Panels.Move(litem_idx, ritem_idx)

    def reorder_beforeall(self, item_name):
        """Reorder and place item_name before all others.

        Args:
            item_name (str): name of component to be moved
        """
        # FIXME: verify docs description is correct
        apiobj = self.get_rvtapi_object()
        litem_idx = None
        if hasattr(apiobj, 'Panels'):
            for item in apiobj.Panels:
                if item.Source.AutomationName == item_name:
                    litem_idx = apiobj.Panels.IndexOf(item)
            if litem_idx:
                apiobj.Panels.Move(litem_idx, 0)

    def reorder_after(self, item_name, ritem_name):
        """Reorder and place item_name after ritem_name.

        Args:
            item_name (str): name of component to be moved
            ritem_name (str): name of component that should be on the left
        """
        apiobj = self.get_rvtapi_object()
        litem_idx = ritem_idx = None
        if hasattr(apiobj, 'Panels'):
            for item in apiobj.Panels:
                if item.Source.AutomationName == item_name:
                    litem_idx = apiobj.Panels.IndexOf(item)
                elif item.Source.AutomationName == ritem_name:
                    ritem_idx = apiobj.Panels.IndexOf(item)
            if litem_idx and ritem_idx:
                if litem_idx < ritem_idx:
                    apiobj.Panels.Move(litem_idx, ritem_idx)
                elif litem_idx > ritem_idx:
                    apiobj.Panels.Move(litem_idx, ritem_idx + 1)

    def reorder_afterall(self, item_name):
        """Reorder and place item_name after all others.

        Args:
            item_name (str): name of component to be moved
        """
        apiobj = self.get_rvtapi_object()
        litem_idx = None
        if hasattr(apiobj, 'Panels'):
            for item in apiobj.Panels:
                if item.Source.AutomationName == item_name:
                    litem_idx = apiobj.Panels.IndexOf(item)
            if litem_idx:
                max_idx = len(apiobj.Panels) - 1
                apiobj.Panels.Move(litem_idx, max_idx)


# Classes holding existing native ui elements
# (These elements are native and can not be modified) --------------------------
class GenericRevitNativeUIContainer(GenericPyRevitUIContainer):
    """Common base type for native Revit API UI containers."""
    def __init__(self):
        GenericPyRevitUIContainer.__init__(self)

    @staticmethod
    def is_native():
        """Is this container generated by pyRevit or is native."""
        return True

    def activate(self):
        """Activate this container in ui.

        Under current implementation, raises PyRevitUIError exception as
        native Revit API UI components should not be changed.
        """
        return self.deactivate()

    def deactivate(self):
        """Deactivate this container in ui.

        Under current implementation, raises PyRevitUIError exception as
        native Revit API UI components should not be changed.
        """
        raise PyRevitUIError('Can not de/activate native item: {}'
                             .format(self))


class RevitNativeRibbonButton(GenericRevitNativeUIContainer):
    """Revit API UI native ribbon button."""
    def __init__(self, adwnd_ribbon_button):
        GenericRevitNativeUIContainer.__init__(self)

        self.name = \
            safe_strtype(adwnd_ribbon_button.AutomationName)\
            .replace('\r\n', ' ')
        self._rvtapi_object = adwnd_ribbon_button


class RevitNativeRibbonGroupItem(GenericRevitNativeUIContainer):
    """Revit API UI native ribbon button."""
    def __init__(self, adwnd_ribbon_item):
        GenericRevitNativeUIContainer.__init__(self)

        self.name = adwnd_ribbon_item.Source.Title
        self._rvtapi_object = adwnd_ribbon_item

        # finding children on this button group
        for adwnd_ribbon_button in adwnd_ribbon_item.Items:
            self._add_component(RevitNativeRibbonButton(adwnd_ribbon_button))

    def button(self, name):
        """Get button item with given name.

        Args:
            name (str): name of button item to find

        Returns:
            (RevitNativeRibbonButton): button object if found
        """
        return super(RevitNativeRibbonGroupItem, self)._get_component(name)


class RevitNativeRibbonPanel(GenericRevitNativeUIContainer):
    """Revit API UI native ribbon button."""
    def __init__(self, adwnd_ribbon_panel):
        GenericRevitNativeUIContainer.__init__(self)

        self.name = adwnd_ribbon_panel.Source.Title
        self._rvtapi_object = adwnd_ribbon_panel

        all_adwnd_ribbon_items = []
        # getting a list of existing items under this panel
        # RibbonFoldPanel items are not visible. they automatically fold
        # buttons into stack on revit ui resize since RibbonFoldPanel are
        # not visible it does not make sense to create objects for them.
        # This pre-cleaner loop, finds the RibbonFoldPanel items and
        # adds the children to the main list
        for adwnd_ribbon_item in adwnd_ribbon_panel.Source.Items:
            if isinstance(adwnd_ribbon_item, AdWindows.RibbonFoldPanel):
                try:
                    for sub_rvtapi_item in adwnd_ribbon_item.Items:
                        all_adwnd_ribbon_items.append(sub_rvtapi_item)
                except Exception as append_err:
                    mlogger.debug('Can not get RibbonFoldPanel children: %s '
                                  '| %s', adwnd_ribbon_item, append_err)
            else:
                all_adwnd_ribbon_items.append(adwnd_ribbon_item)

        # processing the panel slideout for exising ribbon items
        for adwnd_slideout_item \
                in adwnd_ribbon_panel.Source.SlideOutPanelItemsView:
            all_adwnd_ribbon_items.append(adwnd_slideout_item)

        # processing the cleaned children list and
        # creating pyRevit native ribbon objects
        for adwnd_ribbon_item in all_adwnd_ribbon_items:
            try:
                if isinstance(adwnd_ribbon_item,
                              AdWindows.RibbonButton) \
                        or isinstance(adwnd_ribbon_item,
                                      AdWindows.RibbonToggleButton):
                    self._add_component(
                        RevitNativeRibbonButton(adwnd_ribbon_item))
                elif isinstance(adwnd_ribbon_item,
                                AdWindows.RibbonSplitButton):
                    self._add_component(
                        RevitNativeRibbonGroupItem(adwnd_ribbon_item))

            except Exception as append_err:
                mlogger.debug('Can not create native ribbon item: %s '
                              '| %s', adwnd_ribbon_item, append_err)

    def ribbon_item(self, item_name):
        """Get panel item with given name.

        Args:
            item_name (str): name of panel item to find

        Returns:
            (object):
                panel item if found, could be :obj:`RevitNativeRibbonButton`
                or :obj:`RevitNativeRibbonGroupItem`
        """
        return super(RevitNativeRibbonPanel, self)._get_component(item_name)


class RevitNativeRibbonTab(GenericRevitNativeUIContainer):
    """Revit API UI native ribbon tab."""
    def __init__(self, adwnd_ribbon_tab):
        GenericRevitNativeUIContainer.__init__(self)

        self.name = adwnd_ribbon_tab.Title
        self._rvtapi_object = adwnd_ribbon_tab

        # getting a list of existing panels under this tab
        try:
            for adwnd_ribbon_panel in adwnd_ribbon_tab.Panels:
                # only listing visible panels
                if adwnd_ribbon_panel.IsVisible:
                    self._add_component(
                        RevitNativeRibbonPanel(adwnd_ribbon_panel)
                    )
        except Exception as append_err:
            mlogger.debug('Can not get native panels for this native tab: %s '
                          '| %s', adwnd_ribbon_tab, append_err)

    def ribbon_panel(self, panel_name):
        """Get panel with given name.

        Args:
            panel_name (str): name of panel to find

        Returns:
            (RevitNativeRibbonPanel): panel if found
        """
        return super(RevitNativeRibbonTab, self)._get_component(panel_name)

    @staticmethod
    def is_pyrevit_tab():
        """Is this tab generated by pyRevit."""
        return False


# Classes holding non-native ui elements --------------------------------------
class _PyRevitSeparator(GenericPyRevitUIContainer):
    def __init__(self):
        GenericPyRevitUIContainer.__init__(self)
        self.name = coreutils.new_uuid()
        self.itemdata_mode = True


class _PyRevitRibbonButton(GenericPyRevitUIContainer):
    def __init__(self, ribbon_button):
        GenericPyRevitUIContainer.__init__(self)

        self.name = ribbon_button.Name
        self._rvtapi_object = ribbon_button

        # when container is in itemdata_mode, self._rvtapi_object is a
        # RibbonItemData and not an actual ui item a sunsequent call to
        # create_data_items will create ui for RibbonItemData objects
        self.itemdata_mode = isinstance(self._rvtapi_object,
                                        UI.RibbonItemData)

        self.ui_title = self.name
        if not self.itemdata_mode:
            self.ui_title = self._rvtapi_object.ItemText

        self.tooltip_image = self.tooltip_video = None

    def set_rvtapi_object(self, rvtapi_obj):
        GenericPyRevitUIContainer.set_rvtapi_object(self, rvtapi_obj)
        # update the ui title for the newly added rvtapi_obj
        self._rvtapi_object.ItemText = self.ui_title

    def set_icon(self, icon_file, icon_size=ICON_MEDIUM):
        try:
            button_icon = ButtonIcons(icon_file)
            rvtapi_obj = self.get_rvtapi_object()
            rvtapi_obj.Image = button_icon.small_bitmap
            if icon_size == ICON_LARGE:
                rvtapi_obj.LargeImage = button_icon.large_bitmap
            else:
                rvtapi_obj.LargeImage = button_icon.medium_bitmap
            self._dirty = True
        except Exception as icon_err:
            raise PyRevitUIError('Error in applying icon to button > {} : {}'
                                 .format(icon_file, icon_err))

    def set_tooltip(self, tooltip):
        try:
            if tooltip:
                self.get_rvtapi_object().ToolTip = tooltip
            else:
                adwindows_obj = self.get_adwindows_object()
                if adwindows_obj and adwindows_obj.ToolTip:
                    adwindows_obj.ToolTip.Content = None
            self._dirty = True
        except Exception as tooltip_err:
            raise PyRevitUIError('Item does not have tooltip property: {}'
                                 .format(tooltip_err))

    def set_tooltip_ext(self, tooltip_ext):
        try:
            if tooltip_ext:
                self.get_rvtapi_object().LongDescription = tooltip_ext
            else:
                adwindows_obj = self.get_adwindows_object()
                if adwindows_obj and adwindows_obj.ToolTip:
                    adwindows_obj.ToolTip.ExpandedContent = None
            self._dirty = True
        except Exception as tooltip_err:
            raise PyRevitUIError('Item does not have extended '
                                 'tooltip property: {}'.format(tooltip_err))

    def set_tooltip_image(self, tooltip_image):
        try:
            adwindows_obj = self.get_adwindows_object()
            if adwindows_obj:
                exToolTip = self.get_rvtapi_object().ToolTip
                if not isinstance(exToolTip, str):
                    exToolTip = None
                adwindows_obj.ToolTip = AdWindows.RibbonToolTip()
                adwindows_obj.ToolTip.Title = self.ui_title
                adwindows_obj.ToolTip.Content = exToolTip
                _StackPanel = System.Windows.Controls.StackPanel()
                _image = System.Windows.Controls.Image()
                _image.Source = load_bitmapimage(tooltip_image)
                _StackPanel.Children.Add(_image)
                adwindows_obj.ToolTip.ExpandedContent = _StackPanel
                adwindows_obj.ResolveToolTip()
            else:
                self.tooltip_image = tooltip_image
        except Exception as ttimage_err:
            raise PyRevitUIError('Error setting tooltip image {} | {} '
                                 .format(tooltip_image, ttimage_err))

    def set_tooltip_video(self, tooltip_video):
        try:
            adwindows_obj = self.get_adwindows_object()
            if adwindows_obj:
                exToolTip = self.get_rvtapi_object().ToolTip
                if not isinstance(exToolTip, str):
                    exToolTip = None
                adwindows_obj.ToolTip = AdWindows.RibbonToolTip()
                adwindows_obj.ToolTip.Title = self.ui_title
                adwindows_obj.ToolTip.Content = exToolTip
                _StackPanel = System.Windows.Controls.StackPanel()
                _video = System.Windows.Controls.MediaElement()
                _video.Source = Uri(tooltip_video)
                _video.LoadedBehavior = System.Windows.Controls.MediaState.Manual
                _video.UnloadedBehavior = System.Windows.Controls.MediaState.Manual

                def on_media_ended(sender, args):
                    sender.Position = System.TimeSpan.Zero
                    sender.Play()
                _video.MediaEnded += on_media_ended

                def on_loaded(sender, args):
                    sender.Play()
                _video.Loaded += on_loaded
                _StackPanel.Children.Add(_video)
                adwindows_obj.ToolTip.ExpandedContent = _StackPanel
                adwindows_obj.ResolveToolTip()
            else:
                self.tooltip_video = tooltip_video
        except Exception as ttvideo_err:
            raise PyRevitUIError('Error setting tooltip video {} | {} '
                                 .format(tooltip_video, ttvideo_err))

    def set_tooltip_media(self, tooltip_media):
        if tooltip_media.endswith(DEFAULT_TOOLTIP_IMAGE_FORMAT):
            self.set_tooltip_image(tooltip_media)
        elif tooltip_media.endswith(DEFAULT_TOOLTIP_VIDEO_FORMAT):
            self.set_tooltip_video(tooltip_media)

    def reset_highlights(self):
        if hasattr(AdInternal.Windows, 'HighlightMode'):
            adwindows_obj = self.get_adwindows_object()
            if adwindows_obj:
                adwindows_obj.Highlight = \
                    coreutils.get_enum_none(AdInternal.Windows.HighlightMode)

    def highlight_as_new(self):
        if hasattr(AdInternal.Windows, 'HighlightMode'):
            adwindows_obj = self.get_adwindows_object()
            if adwindows_obj:
                adwindows_obj.Highlight = \
                    AdInternal.Windows.HighlightMode.New

    def highlight_as_updated(self):
        if hasattr(AdInternal.Windows, 'HighlightMode'):
            adwindows_obj = self.get_adwindows_object()
            if adwindows_obj:
                adwindows_obj.Highlight = \
                    AdInternal.Windows.HighlightMode.Updated

    def process_deferred(self):
        GenericPyRevitUIContainer.process_deferred(self)

        try:
            if self.tooltip_image:
                self.set_tooltip_image(self.tooltip_image)
        except Exception as ttvideo_err:
            raise PyRevitUIError('Error setting deffered tooltip image {} | {} '
                                 .format(self.tooltip_video, ttvideo_err))

        try:
            if self.tooltip_video:
                self.set_tooltip_video(self.tooltip_video)
        except Exception as ttvideo_err:
            raise PyRevitUIError('Error setting deffered tooltip video {} | {} '
                                 .format(self.tooltip_video, ttvideo_err))

    def get_contexthelp(self):
        return self.get_rvtapi_object().GetContextualHelp()

    def set_contexthelp(self, ctxhelpurl):
        if ctxhelpurl:
            ch = UI.ContextualHelp(UI.ContextualHelpType.Url, ctxhelpurl)
            self.get_rvtapi_object().SetContextualHelp(ch)

    def set_title(self, ui_title):
        if self.itemdata_mode:
            self.ui_title = ui_title
            self._dirty = True
        else:
            self._rvtapi_object.ItemText = self.ui_title = ui_title
            self._dirty = True

    def get_title(self):
        if self.itemdata_mode:
            return self.ui_title
        else:
            return self._rvtapi_object.ItemText

    def get_control_id(self):
        adwindows_obj = self.get_adwindows_object()
        if adwindows_obj and hasattr(adwindows_obj, 'Id'):
            return getattr(adwindows_obj, 'Id', '')

    @property
    def assembly_name(self):
        return self._rvtapi_object.AssemblyName

    @property
    def class_name(self):
        return self._rvtapi_object.ClassName

    @property
    def availability_class_name(self):
        return self._rvtapi_object.AvailabilityClassName


class _PyRevitRibbonGroupItem(GenericPyRevitUIContainer):

    button = GenericPyRevitUIContainer._get_component

    def __init__(self, ribbon_item):
        GenericPyRevitUIContainer.__init__(self)

        self.name = ribbon_item.Name
        self._rvtapi_object = ribbon_item

        # when container is in itemdata_mode, self._rvtapi_object is a
        # RibbonItemData and not an actual ui item when container is in
        # itemdata_mode, only the necessary RibbonItemData objects will
        # be created for children a sunsequent call to create_data_items
        # will create ui for RibbonItemData objects
        self.itemdata_mode = isinstance(self._rvtapi_object,
                                        UI.RibbonItemData)

        # if button group shows the active button icon, then the child
        # buttons need to have large icons
        self._use_active_item_icon = self.is_splitbutton()

        # by default the last item used, stays on top as the default button
        self._sync_with_cur_item = True

        # getting a list of existing items under this item group.
        if not self.itemdata_mode:
            for revit_button in ribbon_item.GetItems():
                # feeding _sub_native_ribbon_items with an instance of
                # _PyRevitRibbonButton for existing buttons
                self._add_component(_PyRevitRibbonButton(revit_button))

    def is_splitbutton(self):
        return isinstance(self._rvtapi_object, UI.SplitButton) \
               or isinstance(self._rvtapi_object, UI.SplitButtonData)

    def set_rvtapi_object(self, rvtapi_obj):
        GenericPyRevitUIContainer.set_rvtapi_object(self, rvtapi_obj)
        if self.is_splitbutton():
            self.get_rvtapi_object().IsSynchronizedWithCurrentItem = \
                self._sync_with_cur_item

    def create_data_items(self):
        # iterate through data items and their associated revit
        # api data objects and create ui objects
        for pyrvt_ui_item in [x for x in self if x.itemdata_mode]:
            rvtapi_data_obj = pyrvt_ui_item.get_rvtapi_object()

            # create item in ui and get correspoding revit ui objects
            if isinstance(pyrvt_ui_item, _PyRevitRibbonButton):
                rvtapi_ribbon_item = \
                    self.get_rvtapi_object().AddPushButton(rvtapi_data_obj)
                rvtapi_ribbon_item.ItemText = pyrvt_ui_item.get_title()

                # replace data object with the newly create ribbon item
                pyrvt_ui_item.set_rvtapi_object(rvtapi_ribbon_item)

                # extended tooltips (images and videos) can only be applied when
                # the ui element is created
                pyrvt_ui_item.process_deferred()

            elif isinstance(pyrvt_ui_item, _PyRevitSeparator):
                self.get_rvtapi_object().AddSeparator()

        self.itemdata_mode = False

    def sync_with_current_item(self, state):
        try:
            if not self.itemdata_mode:
                self.get_rvtapi_object().IsSynchronizedWithCurrentItem = state
            self._sync_with_cur_item = state
            self._dirty = True
        except Exception as sync_item_err:
            raise PyRevitUIError('Item is not a split button. '
                                 '| {}'.format(sync_item_err))

    def set_icon(self, icon_file, icon_size=ICON_LARGE):
        try:
            button_icon = ButtonIcons(icon_file)
            rvtapi_obj = self.get_rvtapi_object()
            rvtapi_obj.Image = button_icon.small_bitmap
            if icon_size == ICON_LARGE:
                rvtapi_obj.LargeImage = button_icon.large_bitmap
            else:
                rvtapi_obj.LargeImage = button_icon.medium_bitmap
            self._dirty = True
        except Exception as icon_err:
            raise PyRevitUIError('Error in applying icon to button > {} : {}'
                                 .format(icon_file, icon_err))

    def get_contexthelp(self):
        return self.get_rvtapi_object().GetContextualHelp()

    def set_contexthelp(self, ctxhelpurl):
        if ctxhelpurl:
            ch = UI.ContextualHelp(UI.ContextualHelpType.Url, ctxhelpurl)
            self.get_rvtapi_object().SetContextualHelp(ch)

    def reset_highlights(self):
        if hasattr(AdInternal.Windows, 'HighlightMode'):
            adwindows_obj = self.get_adwindows_object()
            if adwindows_obj:
                adwindows_obj.HighlightDropDown = \
                    coreutils.get_enum_none(AdInternal.Windows.HighlightMode)

    def highlight_as_new(self):
        if hasattr(AdInternal.Windows, 'HighlightMode'):
            adwindows_obj = self.get_adwindows_object()
            if adwindows_obj:
                adwindows_obj.HighlightDropDown = \
                    AdInternal.Windows.HighlightMode.New

    def highlight_as_updated(self):
        if hasattr(AdInternal.Windows, 'HighlightMode'):
            adwindows_obj = self.get_adwindows_object()
            if adwindows_obj:
                adwindows_obj.HighlightDropDown = \
                    AdInternal.Windows.HighlightMode.Updated

    def create_push_button(self, button_name, asm_location, class_name,
                           icon_path='',
                           tooltip='', tooltip_ext='', tooltip_media='',
                           ctxhelpurl=None,
                           avail_class_name=None,
                           update_if_exists=False, ui_title=None):
        if self.contains(button_name):
            if update_if_exists:
                existing_item = self._get_component(button_name)
                try:
                    # Assembly and Class info of current active script
                    # button can not be updated.
                    if button_name != EXEC_PARAMS.command_name:
                        rvtapi_obj = existing_item.get_rvtapi_object()
                        rvtapi_obj.AssemblyName = asm_location
                        rvtapi_obj.ClassName = class_name
                        if avail_class_name:
                            existing_item.get_rvtapi_object() \
                                .AvailabilityClassName = avail_class_name
                except Exception as asm_update_err:
                    mlogger.debug('Error updating button asm info: %s '
                                  '| %s', button_name, asm_update_err)

                if not icon_path:
                    mlogger.debug('Icon not set for %s', button_name)
                else:
                    try:
                        # if button group shows the active button icon,
                        # then the child buttons need to have large icons
                        existing_item.set_icon(icon_path,
                                               icon_size=ICON_LARGE
                                               if self._use_active_item_icon
                                               else ICON_MEDIUM)
                    except PyRevitUIError as iconerr:
                        mlogger.error('Error adding icon for %s | %s',
                                      button_name, iconerr)

                existing_item.set_tooltip(tooltip)
                existing_item.set_tooltip_ext(tooltip_ext)
                if tooltip_media:
                    existing_item.set_tooltip_media(tooltip_media)

                # if ctx help on this group matches the existing,
                # update self ctx before changing the existing item ctx help
                self_ctxhelp = self.get_contexthelp()
                ctx_help = existing_item.get_contexthelp()
                if self_ctxhelp and ctx_help \
                        and self_ctxhelp.HelpType == ctx_help.HelpType \
                        and self_ctxhelp.HelpPath == ctx_help.HelpPath:
                    self.set_contexthelp(ctxhelpurl)
                # now change the existing item ctx help
                existing_item.set_contexthelp(ctxhelpurl)

                if ui_title:
                    existing_item.set_title(ui_title)

                existing_item.activate()
                return
            else:
                raise PyRevitUIError('Push button already exits and update '
                                     'is not allowed: {}'.format(button_name))

        mlogger.debug('Parent does not include this button. Creating: %s',
                      button_name)
        try:
            button_data = \
                UI.PushButtonData(button_name,
                                  button_name,
                                  asm_location,
                                  class_name)
            if avail_class_name:
                button_data.AvailabilityClassName = avail_class_name
            if not self.itemdata_mode:
                ribbon_button = \
                    self.get_rvtapi_object().AddPushButton(button_data)
                new_button = _PyRevitRibbonButton(ribbon_button)
            else:
                new_button = _PyRevitRibbonButton(button_data)

            if ui_title:
                new_button.set_title(ui_title)

            if not icon_path:
                mlogger.debug('Icon not set for %s', button_name)
            else:
                mlogger.debug('Creating icon for push button %s from file: %s',
                              button_name, icon_path)
                try:
                    # if button group shows the active button icon,
                    # then the child buttons need to have large icons
                    new_button.set_icon(
                        icon_path,
                        icon_size=ICON_LARGE
                        if self._use_active_item_icon else ICON_MEDIUM)
                except PyRevitUIError as iconerr:
                    mlogger.debug('Error adding icon for %s from %s '
                                  '| %s', button_name, icon_path, iconerr)

            new_button.set_tooltip(tooltip)
            new_button.set_tooltip_ext(tooltip_ext)
            if tooltip_media:
                new_button.set_tooltip_media(tooltip_media)

            new_button.set_contexthelp(ctxhelpurl)
            # if this is the first button being added
            if not self.keys():
                mlogger.debug('Setting ctx help on parent: %s', ctxhelpurl)
                self.set_contexthelp(ctxhelpurl)

            new_button.set_dirty_flag()
            self._add_component(new_button)

        except Exception as create_err:
            raise PyRevitUIError('Can not create button '
                                 '| {}'.format(create_err))

    def add_separator(self):
        if not self.itemdata_mode:
            self.get_rvtapi_object().AddSeparator()
        else:
            sep_cmp = _PyRevitSeparator()
            self._add_component(sep_cmp)
        self._dirty = True


class _PyRevitRibbonPanel(GenericPyRevitUIContainer):

    button = GenericPyRevitUIContainer._get_component
    ribbon_item = GenericPyRevitUIContainer._get_component

    def __init__(self, rvt_ribbon_panel, parent_tab):
        GenericPyRevitUIContainer.__init__(self)

        self.name = rvt_ribbon_panel.Name
        self._rvtapi_object = rvt_ribbon_panel

        self.parent_tab = parent_tab

        # when container is in itemdata_mode, only the necessary
        # RibbonItemData objects will be created for children a sunsequent
        # call to create_data_items will create ui for RibbonItemData objects
        # This is specifically helpful when creating stacks in panels.
        # open_stack and close_stack control this parameter
        self.itemdata_mode = False

        # getting a list of existing panels under this tab
        for revit_ribbon_item in self.get_rvtapi_object().GetItems():
            # feeding _sub_native_ribbon_items with an instance of
            # _PyRevitRibbonGroupItem for existing group items
            # _PyRevitRibbonPanel will find its existing ribbon items internally
            if isinstance(revit_ribbon_item, UI.PulldownButton):
                self._add_component(
                    _PyRevitRibbonGroupItem(revit_ribbon_item))
            elif isinstance(revit_ribbon_item, UI.PushButton):
                self._add_component(_PyRevitRibbonButton(revit_ribbon_item))
            else:
                raise PyRevitUIError('Can not determin ribbon item type: {}'
                                     .format(revit_ribbon_item))

    def get_adwindows_object(self):
        for panel in self.parent_tab.Panels:
            if panel.Source and panel.Source.Title == self.name:
                return panel

    def set_background(self, argb_color):
        panel_adwnd_obj = self.get_adwindows_object()
        color = argb_to_brush(argb_color)
        panel_adwnd_obj.CustomPanelBackground = color
        panel_adwnd_obj.CustomPanelTitleBarBackground = color
        panel_adwnd_obj.CustomSlideOutPanelBackground = color

    def reset_backgrounds(self):
        panel_adwnd_obj = self.get_adwindows_object()
        panel_adwnd_obj.CustomPanelBackground = None
        panel_adwnd_obj.CustomPanelTitleBarBackground = None
        panel_adwnd_obj.CustomSlideOutPanelBackground = None

    def set_panel_background(self, argb_color):
        panel_adwnd_obj = self.get_adwindows_object()
        panel_adwnd_obj.CustomPanelBackground = \
            argb_to_brush(argb_color)

    def set_title_background(self, argb_color):
        panel_adwnd_obj = self.get_adwindows_object()
        panel_adwnd_obj.CustomPanelTitleBarBackground = \
            argb_to_brush(argb_color)

    def set_slideout_background(self, argb_color):
        panel_adwnd_obj = self.get_adwindows_object()
        panel_adwnd_obj.CustomSlideOutPanelBackground = \
            argb_to_brush(argb_color)

    def reset_highlights(self):
        # no highlighting options for panels
        pass

    def highlight_as_new(self):
        # no highlighting options for panels
        pass

    def highlight_as_updated(self):
        # no highlighting options for panels
        pass

    def get_collapse(self):
        panel_adwnd_obj = self.get_adwindows_object()
        return panel_adwnd_obj.IsCollapsed

    def set_collapse(self, state):
        panel_adwnd_obj = self.get_adwindows_object()
        panel_adwnd_obj.IsCollapsed = state

    def open_stack(self):
        self.itemdata_mode = True

    def close_stack(self):
        self._create_data_items()

    def add_separator(self):
        self.get_rvtapi_object().AddSeparator()
        self._dirty = True

    def add_slideout(self):
        try:
            self.get_rvtapi_object().AddSlideOut()
            self._dirty = True
        except Exception as slideout_err:
            raise PyRevitUIError('Error adding slide out: {}'
                                 .format(slideout_err))

    def _create_data_items(self):
        # FIXME: if one item changes in stack and others dont change,
        # button will be created as pushbutton out of stack
        self.itemdata_mode = False

        # get a list of data item names and the
        # associated revit api data objects
        pyrvt_data_item_names = [x.name for x in self if x.itemdata_mode]
        rvtapi_data_objs = [x.get_rvtapi_object()
                            for x in self if x.itemdata_mode]

        # list of newly created revit_api ribbon items
        created_rvtapi_ribbon_items = []

        # create stack items in ui and get correspoding revit ui objects
        data_obj_count = len(rvtapi_data_objs)

        # if there are two or 3 items, create a proper stack
        if data_obj_count == 2 or data_obj_count == 3:
            created_rvtapi_ribbon_items = \
                self.get_rvtapi_object().AddStackedItems(*rvtapi_data_objs)
        # if there is only one item added,
        # add that to panel and forget about stacking
        elif data_obj_count == 1:
            rvtapi_pushbutton = \
                self.get_rvtapi_object().AddItem(*rvtapi_data_objs)
            created_rvtapi_ribbon_items.append(rvtapi_pushbutton)
        # if no items have been added, log the empty stack and return
        elif data_obj_count == 0:
            mlogger.debug('No new items has been added to stack. '
                          'Skipping stack creation.')
        # if none of the above, more than 3 items have been added.
        # Cleanup data item cache and raise an error.
        else:
            for pyrvt_data_item_name in pyrvt_data_item_names:
                self._remove_component(pyrvt_data_item_name)
            raise PyRevitUIError('Can not create stack of {}. '
                                 'Stack can only have 2 or 3 items.'
                                 .format(data_obj_count))

        # now that items are created and revit api objects are ready
        # iterate over the ribbon items and inject revit api objects
        # into the child pyrevit items
        for rvtapi_ribbon_item, pyrvt_data_item_name \
                in zip(created_rvtapi_ribbon_items, pyrvt_data_item_names):
            pyrvt_ui_item = self._get_component(pyrvt_data_item_name)
            # pyrvt_ui_item only had button data info.
            # Now that ui ribbon item has created, update pyrvt_ui_item
            # with corresponding revit api object.
            # .set_rvtapi_object() disables .itemdata_mode since
            # they're no longer data objects
            pyrvt_ui_item.set_rvtapi_object(rvtapi_ribbon_item)

            # extended tooltips (images and videos) can only be applied when
            # the ui element is created
            if isinstance(pyrvt_ui_item, _PyRevitRibbonButton):
                pyrvt_ui_item.process_deferred()

            # if pyrvt_ui_item is a group,
            # create children and update group item data
            if isinstance(pyrvt_ui_item, _PyRevitRibbonGroupItem):
                pyrvt_ui_item.create_data_items()

    def set_dlglauncher(self, dlg_button):
        panel_adwnd_obj = self.get_adwindows_object()
        button_adwnd_obj = dlg_button.get_adwindows_object()
        panel_adwnd_obj.Source.Items.Remove(button_adwnd_obj)
        panel_adwnd_obj.Source.DialogLauncher = button_adwnd_obj
        mlogger.debug('Added panel dialog button %s', dlg_button.name)

    def create_push_button(self, button_name, asm_location, class_name,
                           icon_path='',
                           tooltip='', tooltip_ext='', tooltip_media='',
                           ctxhelpurl=None,
                           avail_class_name=None,
                           update_if_exists=False, ui_title=None):
        if self.contains(button_name):
            if update_if_exists:
                existing_item = self._get_component(button_name)
                try:
                    # Assembly and Class info of current active
                    # script button can not be updated.
                    if button_name != EXEC_PARAMS.command_name:
                        rvtapi_obj = existing_item.get_rvtapi_object()
                        rvtapi_obj.AssemblyName = asm_location
                        rvtapi_obj.ClassName = class_name
                        if avail_class_name:
                            rvtapi_obj.AvailabilityClassName = avail_class_name
                except Exception as asm_update_err:
                    mlogger.debug('Error updating button asm info: %s '
                                  '| %s', button_name, asm_update_err)

                existing_item.set_tooltip(tooltip)
                existing_item.set_tooltip_ext(tooltip_ext)
                if tooltip_media:
                    existing_item.set_tooltip_media(tooltip_media)

                existing_item.set_contexthelp(ctxhelpurl)

                if ui_title:
                    existing_item.set_title(ui_title)

                if not icon_path:
                    mlogger.debug('Icon not set for %s', button_name)
                else:
                    try:
                        existing_item.set_icon(icon_path, icon_size=ICON_LARGE)
                    except PyRevitUIError as iconerr:
                        mlogger.error('Error adding icon for %s '
                                      '| %s', button_name, iconerr)
                existing_item.activate()
            else:
                raise PyRevitUIError('Push button already exits and update '
                                     'is not allowed: {}'.format(button_name))
        else:
            mlogger.debug('Parent does not include this button. Creating: %s',
                          button_name)
            try:
                button_data = \
                    UI.PushButtonData(button_name,
                                      button_name,
                                      asm_location,
                                      class_name)
                if avail_class_name:
                    button_data.AvailabilityClassName = avail_class_name
                if not self.itemdata_mode:
                    ribbon_button = \
                        self.get_rvtapi_object().AddItem(button_data)
                    new_button = _PyRevitRibbonButton(ribbon_button)
                else:
                    new_button = _PyRevitRibbonButton(button_data)

                if ui_title:
                    new_button.set_title(ui_title)

                if not icon_path:
                    mlogger.debug('Parent ui item is a panel and '
                                  'panels don\'t have icons.')
                else:
                    mlogger.debug('Creating icon for push button %s '
                                  'from file: %s', button_name, icon_path)
                    try:
                        new_button.set_icon(icon_path, icon_size=ICON_LARGE)
                    except PyRevitUIError as iconerr:
                        mlogger.error('Error adding icon for %s from %s '
                                      '| %s', button_name, icon_path, iconerr)

                new_button.set_tooltip(tooltip)
                new_button.set_tooltip_ext(tooltip_ext)
                if tooltip_media:
                    new_button.set_tooltip_media(tooltip_media)

                new_button.set_contexthelp(ctxhelpurl)

                new_button.set_dirty_flag()
                self._add_component(new_button)

            except Exception as create_err:
                raise PyRevitUIError('Can not create button | {}'
                                     .format(create_err))

    def _create_button_group(self, pulldowndata_type, item_name, icon_path,
                             update_if_exists=False):
        if self.contains(item_name):
            if update_if_exists:
                exiting_item = self._get_component(item_name)
                exiting_item.activate()
                if icon_path:
                    exiting_item.set_icon(icon_path)
            else:
                raise PyRevitUIError('Pull down button already exits and '
                                     'update is not allowed: {}'
                                     .format(item_name))
        else:
            mlogger.debug('Panel does not include this pull down button. '
                          'Creating: %s', item_name)
            try:
                # creating pull down button data and add to child list
                pdbutton_data = pulldowndata_type(item_name, item_name)
                if not self.itemdata_mode:
                    mlogger.debug('Creating pull down button: %s in %s',
                                  item_name, self)
                    new_push_button = \
                        self.get_rvtapi_object().AddItem(pdbutton_data)
                    pyrvt_pdbutton = _PyRevitRibbonGroupItem(new_push_button)
                    try:
                        pyrvt_pdbutton.set_icon(icon_path)
                    except PyRevitUIError as iconerr:
                        mlogger.debug('Error adding icon for %s from %s '
                                      '| %s', item_name, icon_path, iconerr)
                else:
                    mlogger.debug('Creating pull down button under stack: '
                                  '%s in %s', item_name, self)
                    pyrvt_pdbutton = _PyRevitRibbonGroupItem(pdbutton_data)
                    try:
                        pyrvt_pdbutton.set_icon(icon_path)
                    except PyRevitUIError as iconerr:
                        mlogger.debug('Error adding icon for %s from %s '
                                      '| %s', item_name, icon_path, iconerr)

                pyrvt_pdbutton.set_dirty_flag()
                self._add_component(pyrvt_pdbutton)

            except Exception as button_err:
                raise PyRevitUIError('Can not create pull down button: {}'
                                     .format(button_err))

    def create_pulldown_button(self, item_name, icon_path,
                               update_if_exists=False):
        self._create_button_group(UI.PulldownButtonData, item_name, icon_path,
                                  update_if_exists)

    def create_split_button(self, item_name, icon_path,
                            update_if_exists=False):
        if self.itemdata_mode and HOST_APP.is_older_than('2017'):
            raise PyRevitUIError('Revits earlier than 2017 do not support '
                                 'split buttons in a stack.')
        else:
            self._create_button_group(UI.SplitButtonData, item_name, icon_path,
                                      update_if_exists)
            self.ribbon_item(item_name).sync_with_current_item(True)

    def create_splitpush_button(self, item_name, icon_path,
                                update_if_exists=False):
        if self.itemdata_mode and HOST_APP.is_older_than('2017'):
            raise PyRevitUIError('Revits earlier than 2017 do not support '
                                 'split buttons in a stack.')
        else:
            self._create_button_group(UI.SplitButtonData, item_name, icon_path,
                                      update_if_exists)
            self.ribbon_item(item_name).sync_with_current_item(False)

    def create_panel_push_button(self, button_name, asm_location, class_name,
                                 tooltip='', tooltip_ext='', tooltip_media='',
                                 ctxhelpurl=None,
                                 avail_class_name=None,
                                 update_if_exists=False,
                                 ui_title=None):
        self.create_push_button(button_name=button_name,
                                asm_location=asm_location,
                                class_name=class_name,
                                icon_path=None,
                                tooltip=tooltip,
                                tooltip_ext=tooltip_ext,
                                tooltip_media=tooltip_media,
                                ctxhelpurl=ctxhelpurl,
                                avail_class_name=avail_class_name,
                                update_if_exists=update_if_exists,
                                ui_title=ui_title)
        self.set_dlglauncher(self.button(button_name))



class _PyRevitRibbonTab(GenericPyRevitUIContainer):
    ribbon_panel = GenericPyRevitUIContainer._get_component

    def __init__(self, revit_ribbon_tab, is_pyrvt_tab=False):
        GenericPyRevitUIContainer.__init__(self)

        self.name = revit_ribbon_tab.Title
        self._rvtapi_object = revit_ribbon_tab

        # is this tab created by pyrevit.revitui?
        if is_pyrvt_tab:
            self._rvtapi_object.Tag = PYREVIT_TAB_IDENTIFIER

        # getting a list of existing panels under this tab
        try:
            for revit_ui_panel in HOST_APP.uiapp.GetRibbonPanels(self.name):
                # feeding _sub_pyrvt_ribbon_panels with an instance of
                # _PyRevitRibbonPanel for existing panels _PyRevitRibbonPanel
                # will find its existing ribbon items internally
                new_pyrvt_panel = _PyRevitRibbonPanel(revit_ui_panel,
                                                      self._rvtapi_object)
                self._add_component(new_pyrvt_panel)
        except:
            # if .GetRibbonPanels fails, this tab is an existing native tab
            raise PyRevitUIError('Can not get panels for this tab: {}'
                                 .format(self._rvtapi_object))

    def get_adwindows_object(self):
        return self.get_rvtapi_object()

    def reset_highlights(self):
        # no highlighting options for tabs
        pass

    def highlight_as_new(self):
        # no highlighting options for tabs
        pass

    def highlight_as_updated(self):
        # no highlighting options for tabs
        pass

    @staticmethod
    def check_pyrevit_tab(revit_ui_tab):
        return hasattr(revit_ui_tab, 'Tag') \
               and revit_ui_tab.Tag == PYREVIT_TAB_IDENTIFIER

    def is_pyrevit_tab(self):
        return self.get_rvtapi_object().Tag == PYREVIT_TAB_IDENTIFIER

    def update_name(self, new_name):
        self.get_rvtapi_object().Title = new_name

    def create_ribbon_panel(self, panel_name, update_if_exists=False):
        """Create ribbon panel (RevitUI.RibbonPanel) from panel_name."""
        if self.contains(panel_name):
            if update_if_exists:
                exiting_pyrvt_panel = self._get_component(panel_name)
                exiting_pyrvt_panel.activate()
            else:
                raise PyRevitUIError('RibbonPanel already exits and update '
                                     'is not allowed: {}'.format(panel_name))
        else:
            try:
                # creating panel in tab
                ribbon_panel = \
                    HOST_APP.uiapp.CreateRibbonPanel(self.name, panel_name)
                # creating _PyRevitRibbonPanel object and
                # add new panel to list of current panels
                pyrvt_ribbon_panel = _PyRevitRibbonPanel(ribbon_panel,
                                                         self._rvtapi_object)
                pyrvt_ribbon_panel.set_dirty_flag()
                self._add_component(pyrvt_ribbon_panel)

            except Exception as panel_err:
                raise PyRevitUIError('Can not create panel: {}'
                                     .format(panel_err))


class _PyRevitUI(GenericPyRevitUIContainer):
    """Captures the existing ui state and elements at creation."""

    ribbon_tab = GenericPyRevitUIContainer._get_component

    def __init__(self, all_native=False):
        GenericPyRevitUIContainer.__init__(self)

        # Revit does not have any method to get a list of current tabs.
        # Getting a list of current tabs using adwindows.dll methods
        # Iterating over tabs,
        # because ComponentManager.Ribbon.Tabs.FindTab(tab.name)
        # does not return invisible tabs
        for revit_ui_tab in AdWindows.ComponentManager.Ribbon.Tabs:
            # feeding self._sub_pyrvt_ribbon_tabs with an instance of
            # _PyRevitRibbonTab or RevitNativeRibbonTab for each existing
            # tab. _PyRevitRibbonTab or RevitNativeRibbonTab will find
            # their existing panels only listing visible tabs
            # (there might be tabs with identical names
            # e.g. there are two Annotate tabs. They are activated as
            # neccessary per context but need to add inactive/invisible
            # pyrevit tabs (PYREVIT_TAB_IDENTIFIER) anyway.
            # if revit_ui_tab.IsVisible
            try:
                if not all_native \
                        and _PyRevitRibbonTab.check_pyrevit_tab(revit_ui_tab):
                    new_pyrvt_tab = _PyRevitRibbonTab(revit_ui_tab)
                else:
                    new_pyrvt_tab = RevitNativeRibbonTab(revit_ui_tab)
                self._add_component(new_pyrvt_tab)
                mlogger.debug('Tab added to the list of tabs: %s',
                              new_pyrvt_tab.name)
            except PyRevitUIError:
                # if _PyRevitRibbonTab(revit_ui_tab) fails,
                # Revit restricts access to its panels RevitNativeRibbonTab
                # uses a different method to access the panels
                # to interact with existing native ui
                new_pyrvt_tab = RevitNativeRibbonTab(revit_ui_tab)
                self._add_component(new_pyrvt_tab)
                mlogger.debug('Native tab added to the list of tabs: %s',
                              new_pyrvt_tab.name)

    def get_adwindows_ribbon_control(self):
        return AdWindows.ComponentManager.Ribbon

    @staticmethod
    def toggle_ribbon_updator(
            state,
            flow_direction=Windows.FlowDirection.LeftToRight):
        # cancel out the ribbon updator from previous runtime version
        current_ribbon_updator = \
            envvars.get_pyrevit_env_var(envvars.RIBBONUPDATOR_ENVVAR)
        if current_ribbon_updator:
            current_ribbon_updator.StopUpdatingRibbon()

        # reset env var
        envvars.set_pyrevit_env_var(envvars.RIBBONUPDATOR_ENVVAR, None)
        if state:
            # start or stop the ribbon updator
            panel_set = None
            try:
                main_wnd = ui.get_mainwindow()
                ribbon_root_type = ui.get_ribbon_roottype()
                panel_set = \
                    main_wnd.FindFirstChild[ribbon_root_type](main_wnd)
            except Exception as raex:
                mlogger.error('Error activating ribbon updator. | %s', raex)
                return

            if panel_set:
                types.RibbonEventUtils.StartUpdatingRibbon(
                    panelSet=panel_set,
                    flowDir=flow_direction,
                    tagTag=PYREVIT_TAB_IDENTIFIER
                )
                # set the new colorizer
                envvars.set_pyrevit_env_var(
                    envvars.RIBBONUPDATOR_ENVVAR,
                    types.RibbonEventUtils
                    )

    def set_RTL_flow(self):
        _PyRevitUI.toggle_ribbon_updator(
            state=True,
            flow_direction=Windows.FlowDirection.RightToLeft
            )

    def set_LTR_flow(self):
        # default is LTR, make sure any existing is stopped
        _PyRevitUI.toggle_ribbon_updator(state=False)

    def unset_RTL_flow(self):
        _PyRevitUI.toggle_ribbon_updator(state=False)

    def unset_LTR_flow(self):
        # default is LTR, make sure any existing is stopped
        _PyRevitUI.toggle_ribbon_updator(state=False)

    def get_pyrevit_tabs(self):
        return [tab for tab in self if tab.is_pyrevit_tab()]

    def create_ribbon_tab(self, tab_name, update_if_exists=False):
        if self.contains(tab_name):
            if update_if_exists:
                existing_pyrvt_tab = self._get_component(tab_name)
                existing_pyrvt_tab.activate()
            else:
                raise PyRevitUIError('RibbonTab already exits and update is '
                                     'not allowed: {}'.format(tab_name))
        else:
            try:
                # creating tab in Revit ui
                HOST_APP.uiapp.CreateRibbonTab(tab_name)
                # HOST_APP.uiapp.CreateRibbonTab() does
                # not return the created tab object.
                # so find the tab object in exiting ui
                revit_tab_ctrl = None
                for exiting_rvt_ribbon_tab in \
                        AdWindows.ComponentManager.Ribbon.Tabs:
                    if exiting_rvt_ribbon_tab.Title == tab_name:
                        revit_tab_ctrl = exiting_rvt_ribbon_tab

                # create _PyRevitRibbonTab object with
                # the recovered RibbonTab object
                # and add new _PyRevitRibbonTab to list of current tabs
                if revit_tab_ctrl:
                    pyrvt_ribbon_tab = _PyRevitRibbonTab(revit_tab_ctrl,
                                                         is_pyrvt_tab=True)
                    pyrvt_ribbon_tab.set_dirty_flag()
                    self._add_component(pyrvt_ribbon_tab)
                else:
                    raise PyRevitUIError('Tab created but can not '
                                         'be obtained from ui.')

            except Exception as tab_create_err:
                raise PyRevitUIError('Can not create tab: {}'
                                     .format(tab_create_err))


# Public function to return an instance of _PyRevitUI which is used
# to interact with current ui --------------------------------------------------
def get_current_ui(all_native=False):
    """Revit UI Wrapper class for interacting with current pyRevit UI.

    Returned class provides min required functionality for user interaction

    Examples:
        ```python
        current_ui = pyrevit.session.current_ui()
        this_script = pyrevit.session.get_this_command()
        current_ui.update_button_icon(this_script, new_icon)
        ```

    Returns:
        (_PyRevitUI): wrapper around active ribbon gui
    """
    return _PyRevitUI(all_native=all_native)


def get_uibutton(command_unique_name):
    """Find and return ribbon ui button with given unique id.

    Args:
        command_unique_name (str): unique id of pyRevit command

    Returns:
        (_PyRevitRibbonButton): ui button wrapper object
    """
    # FIXME: verify return type
    pyrvt_tabs = get_current_ui().get_pyrevit_tabs()
    for tab in pyrvt_tabs:
        button = tab.find_child(command_unique_name)
        if button:
            return button
    return None



import traceback
from rpw import revit, DB
from rpw.base import BaseObjectWrapper
from rpw.exceptions import RpwException
from rpw.utils.logger import logger


class Transaction(BaseObjectWrapper):
    """
    Simplifies transactions by applying ``Transaction.Start()`` and
    ``Transaction.Commit()`` before and after the context.
    Automatically rolls back if exception is raised.

    >>> from rpw import db
    >>> with db.Transaction('Move Wall'):
    >>>     wall.DoSomething()

    >>> with db.Transaction('Move Wall') as t:
    >>>     wall.DoSomething()
    >>>     assert t.HasStarted() is True
    >>> assert t.HasEnded() is True

    Wrapped Element:
        self._revit_object = `Revit.DB.Transaction`

    """

    _revit_object_class = DB.Transaction

    def __init__(self, name=None, doc=revit.doc):
        if name is None:
            name = 'RPW Transaction'
        super(Transaction, self).__init__(DB.Transaction(doc, name))
        self.transaction = self._revit_object

    def __enter__(self):
        self.transaction.Start()
        return self

    def __exit__(self, exception, exception_msg, tb):
        if exception:
            self.transaction.RollBack()
            logger.error('Error in Transaction Context: has rolled back.')
            # traceback.print_tb(tb)
            # raise exception # Let exception through
        else:
            try:
                self.transaction.Commit()
            except Exception as exc:
                self.transaction.RollBack()
                logger.error('Error in Transaction Commit: has rolled back.')
                logger.error(exc)
                raise

    @staticmethod
    def ensure(name):
        """ Transaction Manager Decorator

        Decorate any function with ``@Transaction.ensure('Transaction Name')``
        and the funciton will run within a Transaction Context.

        Args:
            name (str): Name of the Transaction

        >>> from rpw import db
        >>> @db.Transaction.ensure('Do Something')
        >>> def set_some_parameter(wall, value):
        >>>     wall.parameters['Comments'].value = value
        >>>
        >>> set_some_parameter(wall, value)
        """
        from functools import wraps

        def wrap(f):
            @wraps(f)
            def wrapped_f(*args, **kwargs):
                with Transaction(name):
                    return_value = f(*args, **kwargs)
                return return_value
            return wrapped_f
        return wrap

    # TODO: Add  __repr__ with Transaction Status
    # TODO: Merge Transaction Status
    # TODO: add check for if transaction is in progress, especially for ensure
    # TODO: add ensure to TransactionGroup


class TransactionGroup(BaseObjectWrapper):
    """
    Similar to Transaction, but for ``DB.Transaction Group``

    >>> from rpw import db
    >>> with db.TransacationGroup('Do Major Task'):
    >>>     with db.Transaction('Do Task'):
    >>>         # Do Stuff

    >>> from rpw import db
    >>> with db.TransacationGroup('Do Major Task', assimilate=False):
    >>>     with db.Transaction('Do Task'):
    >>>         # Do Stuff
    """

    _revit_object_class = DB.TransactionGroup

    def __init__(self, name=None, assimilate=True, doc=revit.doc):
        """
            Args:
                name (str): Name of the Transaction
                assimilate (bool): If assimilates is ``True``,
                    transaction history is `squashed`.
        """
        if name is None:
            name = 'RPW Transaction Group'
        super(TransactionGroup, self).__init__(DB.TransactionGroup(doc, name))
        self.transaction_group = self._revit_object
        self.assimilate = assimilate

    def __enter__(self):
        self.transaction_group.Start()
        return self.transaction_group

    def __exit__(self, exception, exception_msg, tb):
        if exception:
            self.transaction_group.RollBack()
            logger.error('Error in TransactionGroup Context: has rolled back.')
        else:
            try:
                if self.assimilate:
                    self.transaction_group.Assimilate()
                else:
                    self.transaction_group.Commit()
            except Exception as exc:
                self.transaction_group.RollBack()
                logger.error('Error in TransactionGroup Commit: \
                              has rolled back.')
                logger.error(exc)
                raise exc


class DynamoTransaction(object):

    # TODO: Use Dynamo Transaction when HOST is 'Dynamo'

    def __init__(self, name):
        raise NotImplemented
    #     from rpw import TransactionManager
    #     self.transaction = TransactionManager.Instance
    #
    # def __enter__(self):
    #     self.transaction.EnsureInTransaction(doc)
    #
    # def __exit__(self, exception, exception_msg, traceback):
    #     if exception:
    #         pass # self.transaction.RollBack()
    #     else:
    #         try:
    #             self.transaction.TransactionTaskDone()
    #         except:
    #             try:
    #                 self.transaction.ForceCloseTransaction()
    #             except:
    #                 raise RpwException('Failed to complete transaction')



class ConsoleTraceListener(TextWriterTraceListener,IDisposable):
 """
 Directs tracing or debugging output to either the standard output or the standard error stream.

 

 ConsoleTraceListener()

 ConsoleTraceListener(useErrorStream: bool)
 """
 def Close(self):
  """
  Close(self: ConsoleTraceListener)

   Closes the output to the stream specified for this trace listener.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: TextWriterTraceListener,disposing: bool)

   Disposes this System.Diagnostics.TextWriterTraceListener object.

  

   disposing: true to release managed resources; if false,

    System.Diagnostics.TextWriterTraceListener.Dispose(System.Boolean) has no effect.
  """
  pass
 def GetSupportedAttributes(self,*args):
  """
  GetSupportedAttributes(self: TraceListener) -> Array[str]

  

   Gets the custom attributes supported by the trace listener.

   Returns: A string array naming the custom attributes supported by the trace listener,or null if there 

    are no custom attributes.
  """
  pass
 def MemberwiseClone(self,*args):
  """
  MemberwiseClone(self: MarshalByRefObject,cloneIdentity: bool) -> MarshalByRefObject

  

   Creates a shallow copy of the current System.MarshalByRefObject object.

  

   cloneIdentity: false to delete the current System.MarshalByRefObject object's identity,which will cause the 

    object to be assigned a new identity when it is marshaled across a remoting boundary. A value of 

    false is usually appropriate. true to copy the current System.MarshalByRefObject object's 

    identity to its clone,which will cause remoting client calls to be routed to the remote server 

    object.

  

   Returns: A shallow copy of the current System.MarshalByRefObject object.

  MemberwiseClone(self: object) -> object

  

   Creates a shallow copy of the current System.Object.

   Returns: A shallow copy of the current System.Object.
  """
  pass
 def WriteIndent(self,*args):
  """
  WriteIndent(self: TraceListener)

   Writes the indent to the listener you create when you implement this class,and resets the 

    System.Diagnostics.TraceListener.NeedIndent property to false.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,useErrorStream=None):
  """
  __new__(cls: type)

  __new__(cls: type,useErrorStream: bool)
  """
  pass
 NeedIndent=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether to indent the output.



"""





class OSFeature(FeatureSupport,IFeatureSupport):
 """ Provides operating-system specific feature queries. """
 def GetVersionPresent(self,feature):
  """
  GetVersionPresent(self: OSFeature,feature: object) -> Version

  

   Retrieves the version of the specified feature currently available on the system.

  

   feature: The feature whose version is requested,either System.Windows.Forms.OSFeature.LayeredWindows or 

    System.Windows.Forms.OSFeature.Themes.

  

   Returns: A System.Version representing the version of the specified operating system feature currently 

    available on the system; or null if the feature cannot be found.
  """
  pass
 @staticmethod
 def IsPresent(*__args):
  """
  IsPresent(enumVal: SystemParameter) -> bool

  

   Retrieves a value indicating whether the operating system supports the specified feature or 

    metric.

  

  

   enumVal: A System.Windows.Forms.SystemParameter representing the feature to search for.

   Returns: true if the feature is available on the system; otherwise,false.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 Feature=None
 LayeredWindows=None
 Themes=None




# encoding: utf-8
# module System.Windows.Markup calls itself Markup
# from System,Version=4.0.0.0,Culture=neutral,PublicKeyToken=b77a5c561934e089
# by generator 1.145
""" NamespaceTracker represent a CLS namespace. """
# no imports

# no functions
# classes

class ValueSerializerAttribute(Attribute,_Attribute):
 """
 ValueSerializerAttribute(valueSerializerType: Type)

 ValueSerializerAttribute(valueSerializerTypeName: str)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type,valueSerializerType: Type)

  __new__(cls: type,valueSerializerTypeName: str)
  """
  pass
 ValueSerializerType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: ValueSerializerType(self: ValueSerializerAttribute) -> Type



"""

 ValueSerializerTypeName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: ValueSerializerTypeName(self: ValueSerializerAttribute) -> str



"""






class DrawItemState(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies the state of an item that is being drawn.

 

 enum (flags) DrawItemState,values: Checked (8),ComboBoxEdit (4096),Default (32),Disabled (4),Focus (16),Grayed (2),HotLight (64),Inactive (128),NoAccelerator (256),NoFocusRect (512),None (0),Selected (1)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Checked=None
 ComboBoxEdit=None
 Default=None
 Disabled=None
 Focus=None
 Grayed=None
 HotLight=None
 Inactive=None
 NoAccelerator=None
 NoFocusRect=None
 None=None
 Selected=None
 value__=None




"""Helper functions to update info and elements in Revit."""
import os.path as op

from pyrevit import HOST_APP, DOCS
from pyrevit.framework import List
from pyrevit import DB
from pyrevit.revit.db import query
from pyrevit.compat import get_elementid_value_func


def set_name(element, new_name):
    # grab viewname correctly
    if isinstance(element, DB.View):
        if HOST_APP.is_newer_than('2019', or_equal=True):
            element.Name = new_name
        else:
            element.ViewName = new_name
    else:
        element.Name = new_name


def update_sheet_revisions(revisions, sheets=None, state=True, doc=None):
    doc = doc or DOCS.doc
    get_elementid_value = get_elementid_value_func()
    # make sure revisions is a list
    if not isinstance(revisions, list):
        revisions = [revisions]
    updated_sheets = []
    if revisions:
        # get sheets if not available
        for sheet in sheets or query.get_sheets(doc=doc):
            addrevs = set([get_elementid_value(x)
                           for x in sheet.GetAdditionalRevisionIds()])
            for rev in revisions:
                # skip issued revisions
                if not rev.Issued:
                    if state:
                        addrevs.add(get_elementid_value(rev.Id))
                    elif get_elementid_value(rev.Id) in addrevs:
                        addrevs.remove(get_elementid_value(rev.Id))
            rev_elids = [DB.ElementId(x) for x in addrevs]
            sheet.SetAdditionalRevisionIds(List[DB.ElementId](rev_elids))
            updated_sheets.append(sheet)
    return updated_sheets


def update_revision_alphanumeric(token_list, prefix='', postfix='', doc=None):
    doc = doc or DOCS.doc
    alphalist = List[str]()
    for token in token_list:
        alphalist.Add(str(token))
    alpha_cfg = DB.AlphanumericRevisionSettings(alphalist, prefix, postfix)
    rev_cfg = DB.RevisionSettings.GetRevisionSettings(doc)
    rev_cfg.SetAlphanumericRevisionSettings(alpha_cfg)


def update_revision_numeric(starting_int, prefix='', postfix='', doc=None):
    doc = doc or DOCS.doc
    num_cfg = DB.NumericRevisionSettings(starting_int, prefix, postfix)
    rev_cfg = DB.RevisionSettings.GetRevisionSettings(doc)
    rev_cfg.SetNumericRevisionSettings(num_cfg)


def update_revision_numbering(per_sheet=False, doc=None):
    doc = doc or DOCS.doc
    rev_cfg = DB.RevisionSettings.GetRevisionSettings(doc)
    rev_cfg.RevisionNumbering = \
        DB.RevisionNumbering.PerSheet if per_sheet else \
        DB.RevisionNumbering.PerProject


def update_param_value(rvt_param, value):
    if not rvt_param.IsReadOnly:
        if rvt_param.StorageType == DB.StorageType.String:
            rvt_param.Set(str(value) if value else "")
        else:
            rvt_param.SetValueString(str(value))


def toggle_category_visibility(view, subcat, hidden=None):
    if HOST_APP.is_older_than(2018):
        if hidden is None:
            hidden = not view.GetVisibility(subcat.Id)
        view.SetVisibility(subcat.Id, hidden)
    else:
        if hidden is None:
            hidden = not view.GetCategoryHidden(subcat.Id)
        view.SetCategoryHidden(subcat.Id, hidden)


def rename_workset(workset, new_name, doc=None):
    doc = doc or DOCS.doc
    DB.WorksetTable.RenameWorkset(doc, workset.Id, new_name)


def update_linked_keynotes(doc=None):
    doc = doc or DOCS.doc
    ktable = DB.KeynoteTable.GetKeynoteTable(doc)
    ktable.Reload(None)


# https://forum.dynamobim.com/t/load-assemblycodetable-keynotetable/23944/2
def set_keynote_file(keynote_file, doc=None):
    doc = doc or DOCS.doc
    if op.exists(keynote_file):
        mpath = \
            DB.ModelPathUtils.ConvertUserVisiblePathToModelPath(keynote_file)
        keynote_exres = DB.ExternalResourceReference.CreateLocalResource(
            doc,
            DB.ExternalResourceTypes.BuiltInExternalResourceTypes.KeynoteTable,
            mpath,
            DB.PathType.Absolute)
        knote_table = DB.KeynoteTable.GetKeynoteTable(doc)
        knote_table.LoadFrom(keynote_exres, DB.KeyBasedTreeEntriesLoadResults())


def set_crop_region(view, curve_loops):
    """Sets crop region to a view.

    Args:
        view (DB.View): view to change
        curve_loops (list[DB.CurveLoop]): list of curve loops
    """
    if not isinstance(curve_loops, list):
        curve_loops = [curve_loops]

    crop_active_saved = view.CropBoxActive
    view.CropBoxActive = True
    crsm = view.GetCropRegionShapeManager()
    for cloop in curve_loops:
        if HOST_APP.is_newer_than(2015):
            crsm.SetCropShape(cloop)
        else:
            crsm.SetCropRegionShape(cloop)
    view.CropBoxActive = crop_active_saved


def set_active_workset(workset_id, doc=None):
    """Set active workset.

    Args:
        workset_id (DB.WorksetId): target workset id
        doc (DB.Document, optional): target document. defaults to active
    """
    doc = doc or DOCS.doc
    if doc.IsWorkshared:
        workset_table = doc.GetWorksetTable()
        workset_table.SetActiveWorksetId(workset_id)



class ListViewItemMouseHoverEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.ListView.ItemMouseHover event.

 

 ListViewItemMouseHoverEventArgs(item: ListViewItem)
 """
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y] """
  pass
 @staticmethod
 def __new__(self,item):
  """ __new__(cls: type,item: ListViewItem) """
  pass
 Item=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the item the mouse pointer is currently hovering over.



Get: Item(self: ListViewItemMouseHoverEventArgs) -> ListViewItem



"""





class DataGridViewDataErrorContexts(Enum,IComparable,IFormattable,IConvertible):
 """
 Represents the state of a data-bound System.Windows.Forms.DataGridView control when a data error occurred.

 

 enum (flags) DataGridViewDataErrorContexts,values: ClipboardContent (16384),Commit (512),CurrentCellChange (4096),Display (2),Formatting (1),InitialValueRestoration (1024),LeaveControl (2048),Parsing (256),PreferredSize (4),RowDeletion (8),Scroll (8192)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 ClipboardContent=None
 Commit=None
 CurrentCellChange=None
 Display=None
 Formatting=None
 InitialValueRestoration=None
 LeaveControl=None
 Parsing=None
 PreferredSize=None
 RowDeletion=None
 Scroll=None
 value__=None




class NotifyDataErrorValidationRule(ValidationRule):
 """ NotifyDataErrorValidationRule() """
 def Validate(self,value,cultureInfo,owner=None):
  """ Validate(self: NotifyDataErrorValidationRule,value: object,cultureInfo: CultureInfo) -> ValidationResult """
  pass



"""Read and Write STL Binary and ASCII Files."""

import struct
import os.path as op


class STLMesh(object):
    """Container for STL mesh data."""

    def __init__(self):
        self.triangles = []

    def add_triangle(self, normal, vertices):
        """Add a triangle to the mesh.

        Args:
            normal: tuple of 3 floats (nx, ny, nz)
            vertices: list of 3 tuples, each with 3 floats (x, y, z)
        """
        self.triangles.append({"normal": normal, "vertices": vertices})


def load(filepath):
    """Load an STL file (binary or ASCII).

    Args:
        filepath: path to the STL file

    Returns:
        STLMesh object containing the mesh data
    """
    if not op.exists(filepath):
        raise IOError("File not found: {0}".format(filepath))

    # Try to determine if file is binary or ASCII
    with open(filepath, "rb") as f:
        header = f.read(80)
        # ASCII files start with "solid"
        if header.startswith(b"solid") or header.startswith(b"SOLID"):
            # Could be ASCII, but need to verify
            f.seek(0)
            try:
                # Try reading as ASCII
                first_line = f.readline().decode("ascii", errors="strict")
                if first_line.strip().startswith("solid"):
                    f.seek(0)
                    return _load_ascii(f)
            except (UnicodeDecodeError, ValueError):
                pass

        # If not ASCII, load as binary
        f.seek(0)
        return _load_binary(f)


def _load_binary(file_handle):
    """Load a binary STL file.

    Args:
        file_handle: open file handle

    Returns:
        STLMesh object
    """
    mesh = STLMesh()

    # Read 80-byte header (ignored)
    header = file_handle.read(80)

    # Read number of triangles
    triangle_count_data = file_handle.read(4)
    if len(triangle_count_data) < 4:
        raise ValueError("Invalid STL file: cannot read triangle count")

    triangle_count = struct.unpack("<I", triangle_count_data)[0]

    # Read each triangle
    for i in range(triangle_count):
        # Each triangle is 50 bytes:
        # - 3 floats for normal (12 bytes)
        # - 3 vertices * 3 floats each (36 bytes)
        # - 1 uint16 attribute (2 bytes, usually unused)

        triangle_data = file_handle.read(50)
        if len(triangle_data) < 50:
            raise ValueError(
                "Invalid STL file: incomplete triangle data at triangle {0}".format(i)
            )

        # Unpack normal (3 floats)
        normal = struct.unpack("<3f", triangle_data[0:12])

        # Unpack vertices (3 vertices * 3 floats)
        v1 = struct.unpack("<3f", triangle_data[12:24])
        v2 = struct.unpack("<3f", triangle_data[24:36])
        v3 = struct.unpack("<3f", triangle_data[36:48])

        vertices = [v1, v2, v3]

        # Attribute bytes (ignored)
        # attr = struct.unpack('<H', triangle_data[48:50])[0]

        mesh.add_triangle(normal, vertices)

    return mesh


def _load_ascii(file_handle):
    """Load an ASCII STL file.

    Args:
        file_handle: open file handle

    Returns:
        STLMesh object
    """
    mesh = STLMesh()

    current_normal = None
    current_vertices = []
    in_facet = False

    for line in file_handle:
        line = line.strip()

        if not line:
            continue

        # Handle both bytes and str (for Python 2/3 compatibility)
        if isinstance(line, bytes):
            line = line.decode("ascii", errors="ignore")

        line = line.lower()
        parts = line.split()

        if not parts:
            continue

        if parts[0] == "facet" and len(parts) >= 5 and parts[1] == "normal":
            in_facet = True
            current_normal = (float(parts[2]), float(parts[3]), float(parts[4]))
            current_vertices = []

        elif parts[0] == "vertex" and len(parts) >= 4:
            vertex = (float(parts[1]), float(parts[2]), float(parts[3]))
            current_vertices.append(vertex)

        elif parts[0] == "endfacet":
            if in_facet and current_normal and len(current_vertices) == 3:
                mesh.add_triangle(current_normal, current_vertices)
            in_facet = False
            current_normal = None
            current_vertices = []

    return mesh


def dump(outputfile, mesh):
    """Save mesh to binary STL file.

    Args:
        outputfile: path to output file
        mesh: STLMesh object to save
    """
    with open(outputfile, "wb") as f:
        # Write 80-byte header
        header = b"Binary STL file generated by pyRevit"
        header = header + b" " * (80 - len(header))
        f.write(header[:80])

        # Write triangle count
        f.write(struct.pack("<I", len(mesh.triangles)))

        # Write each triangle
        for triangle in mesh.triangles:
            # Write normal
            f.write(struct.pack("<3f", *triangle["normal"]))

            # Write vertices
            for vertex in triangle["vertices"]:
                f.write(struct.pack("<3f", *vertex))

            # Write attribute byte count (0)
            f.write(struct.pack("<H", 0))



"""Perform upgrades between version, e.g. adding a new config parameter."""
#pylint: disable=W0611
import os
import os.path as op

from pyrevit.coreutils import appdata


def upgrade_user_config(user_config):   #pylint: disable=W0613
    """Upgarde user configurations.

    Args:
        user_config (:obj:`pyrevit.userconfig.PyRevitConfig`): config object
    """
    # upgrade value formats
    for section in user_config:
        for option in section:
            setattr(section, option, getattr(section, option))


def remove_leftover_temp_files():
    """4.8.5 had a bug that would create temp files with extension ..bak.

    This cleans them up.
    """
    univ_path = op.dirname(appdata.get_universal_data_file("X", 'bak'))
    if op.exists(univ_path):
        for entry in os.listdir(univ_path):
            if op.isfile(entry) and entry.lower().endswith('..bak'):
                appdata.garbage_data_file(op.join(univ_path, entry))


def upgrade_existing_pyrevit():
    """Upgrade existing pyRevit deployment."""
    remove_leftover_temp_files()



class WebPermission(CodeAccessPermission,IPermission,ISecurityEncodable,IStackWalk,IUnrestrictedPermission):
 """
 Controls rights to access HTTP Internet resources.

 

 WebPermission(state: PermissionState)

 WebPermission()

 WebPermission(access: NetworkAccess,uriRegex: Regex)

 WebPermission(access: NetworkAccess,uriString: str)
 """
 def AddPermission(self,access,*__args):
  """
  AddPermission(self: WebPermission,access: NetworkAccess,uriRegex: Regex)

   Adds the specified URI with the specified access rights to the current System.Net.WebPermission.

  

   access: A NetworkAccess that specifies the access rights that are granted to the URI.

   uriRegex: A regular expression that describes the set of URIs to which access rights are granted.

  AddPermission(self: WebPermission,access: NetworkAccess,uriString: str)

   Adds the specified URI string with the specified access rights to the current 

    System.Net.WebPermission.

  

  

   access: A System.Net.NetworkAccess that specifies the access rights that are granted to the URI.

   uriString: A string that describes the URI to which access rights are granted.
  """
  pass
 def Copy(self):
  """
  Copy(self: WebPermission) -> IPermission

  

   Creates a copy of a System.Net.WebPermission.

   Returns: A new instance of the System.Net.WebPermission class that has the same values as the original.
  """
  pass
 def FromXml(self,securityElement):
  """
  FromXml(self: WebPermission,securityElement: SecurityElement)

   Reconstructs a System.Net.WebPermission from an XML encoding.

  

   securityElement: The XML encoding from which to reconstruct the System.Net.WebPermission.
  """
  pass
 def Intersect(self,target):
  """
  Intersect(self: WebPermission,target: IPermission) -> IPermission

  

   Returns the logical intersection of two System.Net.WebPermission instances.

  

   target: The System.Net.WebPermission to compare with the current instance.

   Returns: A new System.Net.WebPermission that represents the intersection of the current instance and the 

    target parameter. If the intersection is empty,the method returns null.
  """
  pass
 def IsSubsetOf(self,target):
  """
  IsSubsetOf(self: WebPermission,target: IPermission) -> bool

  

   Determines whether the current System.Net.WebPermission is a subset of the specified object.

  

   target: The System.Net.WebPermission to compare to the current System.Net.WebPermission.

   Returns: true if the current instance is a subset of the target parameter; otherwise,false. If the 

    target is null,the method returns true for an empty current permission that is not unrestricted 

    and false otherwise.
  """
  pass
 def IsUnrestricted(self):
  """
  IsUnrestricted(self: WebPermission) -> bool

  

   Checks the overall permission state of the System.Net.WebPermission.

   Returns: true if the System.Net.WebPermission was created with the 

    System.Security.Permissions.PermissionState.UnrestrictedSystem.Security.Permissions.PermissionSta

    te; otherwise,false.
  """
  pass
 def ToXml(self):
  """
  ToXml(self: WebPermission) -> SecurityElement

  

   Creates an XML encoding of a System.Net.WebPermission and its current state.

   Returns: A System.Security.SecurityElement that contains an XML-encoded representation of the 

    System.Net.WebPermission,including state information.
  """
  pass
 def Union(self,target):
  """
  Union(self: WebPermission,target: IPermission) -> IPermission

  

   Returns the logical union between two instances of the System.Net.WebPermission class.

  

   target: The System.Net.WebPermission to combine with the current System.Net.WebPermission.

   Returns: A System.Net.WebPermission that represents the union of the current instance and the target 

    parameter. If either WebPermission is System.Security.Permissions.PermissionState.Unrestricted,

    the method returns a System.Net.WebPermission that is 

    System.Security.Permissions.PermissionState.Unrestricted. If the target is null,the method 

    returns a copy of the current System.Net.WebPermission.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type,state: PermissionState)

  __new__(cls: type)

  __new__(cls: type,access: NetworkAccess,uriRegex: Regex)

  __new__(cls: type,access: NetworkAccess,uriString: str)
  """
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 AcceptList=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """This property returns an enumeration of a single accept permissions held by this System.Net.WebPermission. The possible objects types contained in the returned enumeration are System.String and System.Text.RegularExpressions.Regex.



Get: AcceptList(self: WebPermission) -> IEnumerator



"""

 ConnectList=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """This property returns an enumeration of a single connect permissions held by this System.Net.WebPermission. The possible objects types contained in the returned enumeration are System.String and System.Text.RegularExpressions.Regex.



Get: ConnectList(self: WebPermission) -> IEnumerator



"""





class ObsoleteAttribute(Attribute,_Attribute):
 """
 Marks the program elements that are no longer in use. This class cannot be inherited.

 

 ObsoleteAttribute()

 ObsoleteAttribute(message: str)

 ObsoleteAttribute(message: str,error: bool)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,message=None,error=None):
  """
  __new__(cls: type)

  __new__(cls: type,message: str)

  __new__(cls: type,message: str,error: bool)
  """
  pass
 def __reduce_ex__(self,*args):
  pass
 IsError=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a Boolean value indicating whether the compiler will treat usage of the obsolete program element as an error.



Get: IsError(self: ObsoleteAttribute) -> bool



"""

 Message=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the workaround message,including a description of the alternative program elements.



Get: Message(self: ObsoleteAttribute) -> str



"""





class CultureInfoConverter(TypeConverter):
 """
 Provides a type converter to convert System.Globalization.CultureInfo objects to and from various other representations.

 

 CultureInfoConverter()
 """
 def CanConvertFrom(self,*__args):
  """
  CanConvertFrom(self: CultureInfoConverter,context: ITypeDescriptorContext,sourceType: Type) -> bool

  

   Gets a value indicating whether this converter can convert an object in the given source type to 

    a System.Globalization.CultureInfo using the specified context.

  

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   sourceType: A System.Type that represents the type you wish to convert from.

   Returns: true if this converter can perform the conversion; otherwise,false.
  """
  pass
 def CanConvertTo(self,*__args):
  """
  CanConvertTo(self: CultureInfoConverter,context: ITypeDescriptorContext,destinationType: Type) -> bool

  

   Gets a value indicating whether this converter can convert an object to the given destination 

    type using the context.

  

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   destinationType: A System.Type that represents the type you wish to convert to.

   Returns: true if this converter can perform the conversion; otherwise,false.
  """
  pass
 def ConvertFrom(self,*__args):
  """
  ConvertFrom(self: CultureInfoConverter,context: ITypeDescriptorContext,culture: CultureInfo,value: object) -> object

  

   Converts the specified value object to a System.Globalization.CultureInfo.

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   culture: A System.Globalization.CultureInfo that specifies the culture to which to convert.

   value: The System.Object to convert.

   Returns: An System.Object that represents the converted value.
  """
  pass
 def ConvertTo(self,*__args):
  """
  ConvertTo(self: CultureInfoConverter,context: ITypeDescriptorContext,culture: CultureInfo,value: object,destinationType: Type) -> object

  

   Converts the given value object to the specified destination type.

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   culture: A System.Globalization.CultureInfo that specifies the culture to which to convert.

   value: The System.Object to convert.

   destinationType: The System.Type to convert the value to.

   Returns: An System.Object that represents the converted value.
  """
  pass
 def GetCultureName(self,*args):
  """
  GetCultureName(self: CultureInfoConverter,culture: CultureInfo) -> str

  

   Retrieves the name of the specified culture.

  

   culture: A System.Globalization.CultureInfo that specifies the culture to get the name for.

   Returns: The name of the specified culture.
  """
  pass
 def GetStandardValues(self,context=None):
  """
  GetStandardValues(self: CultureInfoConverter,context: ITypeDescriptorContext) -> StandardValuesCollection

  

   Gets a collection of standard values for a System.Globalization.CultureInfo object using the 

    specified context.

  

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   Returns: A System.ComponentModel.TypeConverter.StandardValuesCollection containing a standard set of 

    valid values,or null if the data type does not support a standard set of values.
  """
  pass
 def GetStandardValuesExclusive(self,context=None):
  """
  GetStandardValuesExclusive(self: CultureInfoConverter,context: ITypeDescriptorContext) -> bool

  

   Gets a value indicating whether the list of standard values returned from 

    System.ComponentModel.CultureInfoConverter.GetStandardValues(System.ComponentModel.ITypeDescripto

    rContext) is an exhaustive list.

  

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   Returns: false because the System.ComponentModel.TypeConverter.StandardValuesCollection returned from 

    System.ComponentModel.CultureInfoConverter.GetStandardValues(System.ComponentModel.ITypeDescripto

    rContext) is not an exhaustive list of possible values (that is,other values are possible). 

    This method never returns true.
  """
  pass
 def GetStandardValuesSupported(self,context=None):
  """
  GetStandardValuesSupported(self: CultureInfoConverter,context: ITypeDescriptorContext) -> bool

  

   Gets a value indicating whether this object supports a standard set of values that can be picked 

    from a list using the specified context.

  

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   Returns: true because 

    System.ComponentModel.CultureInfoConverter.GetStandardValues(System.ComponentModel.ITypeDescripto

    rContext) should be called to find a common set of values the object supports. This method never 

    returns false.
  """
  pass



"""Provide access to Revit API.

Examples:
    ```python
    from pyrevit.api import AdWindows
    ```
"""
import os.path as op

#pylint: disable=E0401,W0611,W0703,C0413
from pyrevit.framework import clr

clr.AddReference('RevitAPI')
clr.AddReference('RevitAPIUI')
clr.AddReference('AdWindows')
clr.AddReference('UIFramework')
clr.AddReference('UIFrameworkServices')

import UIFramework
import UIFrameworkServices

import Autodesk.Internal as AdInternal
import Autodesk.Private as AdPrivate
import Autodesk.Windows as AdWindows

from Autodesk.Revit import Attributes
from Autodesk.Revit import ApplicationServices
from Autodesk.Revit import DB
from Autodesk.Revit import UI
from Autodesk.Revit.DB import ExternalService
from Autodesk.Revit.DB import DirectContext3D


def get_product_serial_number():
    """Return serial number of running host instance.

    Returns:
        (str): Serial number
    """
    return UIFrameworkServices.InfoCenterService.ProductSerialNumber


def is_product_demo():
    """Determine if product is using demo license.

    Returns:
        (bool): True if product is using demo license
    """
    return get_product_serial_number() == '000-00000000'


def is_api_object(data_type):
    """Check if given object belongs to Revit API.

    Args:
        data_type (object): Object to check

    Returns:
        (bool): True if object belongs to Revit API
    """
    if hasattr(data_type, 'GetType'):
        return 'Autodesk.Revit.' in data_type.GetType().Namespace



class UCOMIBindCtx:
 """ Use System.Runtime.InteropServices.ComTypes.BIND_OPTS instead. """
 def EnumObjectParam(self,ppenum):
  """
  EnumObjectParam(self: UCOMIBindCtx) -> UCOMIEnumString

  

   Enumerate the strings which are the keys of the internally-maintained table of contextual object 

    parameters.
  """
  pass
 def GetBindOptions(self,pbindopts):
  """
  GetBindOptions(self: UCOMIBindCtx,pbindopts: BIND_OPTS) -> BIND_OPTS

  

   Return the current binding options stored in this bind context.

  

   pbindopts: A pointer to the structure to receive the binding options.
  """
  pass
 def GetObjectParam(self,pszKey,ppunk):
  """
  GetObjectParam(self: UCOMIBindCtx,pszKey: str) -> object

  

   Lookup the given key in the internally-maintained table of contextual object parameters and 

    return the corresponding object,if one exists.

  

  

   pszKey: The name of the object to search for.
  """
  pass
 def GetRunningObjectTable(self,pprot):
  """
  GetRunningObjectTable(self: UCOMIBindCtx) -> UCOMIRunningObjectTable

  

   Return access to the Running Object Table (ROT) relevant to this binding process.
  """
  pass
 def RegisterObjectBound(self,punk):
  """
  RegisterObjectBound(self: UCOMIBindCtx,punk: object)

   Register the passed object as one of the objects that has been bound during a moniker operation 

    and which should be released when it is complete.

  

  

   punk: The object to register for release.
  """
  pass
 def RegisterObjectParam(self,pszKey,punk):
  """
  RegisterObjectParam(self: UCOMIBindCtx,pszKey: str,punk: object)

   Register the given object pointer under the specified name in the internally-maintained table of 

    object pointers.

  

  

   pszKey: The name to register punk with.

   punk: The object to register.
  """
  pass
 def ReleaseBoundObjects(self):
  """
  ReleaseBoundObjects(self: UCOMIBindCtx)

   Releases all the objects currently registered with the bind context by 

    System.Runtime.InteropServices.UCOMIBindCtx.RegisterObjectBound(System.Object).
  """
  pass
 def RevokeObjectBound(self,punk):
  """
  RevokeObjectBound(self: UCOMIBindCtx,punk: object)

   Removes the object from the set of registered objects that need to be released.

  

   punk: The object to unregister for release.
  """
  pass
 def RevokeObjectParam(self,pszKey):
  """
  RevokeObjectParam(self: UCOMIBindCtx,pszKey: str)

   Revoke the registration of the object currently found under this key in the 

    internally-maintained table of contextual object parameters,if any such key is currently 

    registered.

  

  

   pszKey: The key to unregister.
  """
  pass
 def SetBindOptions(self,pbindopts):
  """
  SetBindOptions(self: UCOMIBindCtx,pbindopts: BIND_OPTS) -> BIND_OPTS

  

   Store in the bind context a block of parameters that will apply to later UCOMIMoniker operations 

    using this bind context.

  

  

   pbindopts: The structure containing the binding options to set.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class DataGridViewButtonColumn(DataGridViewColumn,ICloneable,IDisposable,IComponent):
 """
 Hosts a collection of System.Windows.Forms.DataGridViewButtonCell objects.

 

 DataGridViewButtonColumn()
 """
 def Clone(self):
  """
  Clone(self: DataGridViewButtonColumn) -> object

  

   Creates an exact copy of this column.

   Returns: An System.Object that represents the cloned System.Windows.Forms.DataGridViewButtonColumn.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: DataGridViewColumn,disposing: bool)

   disposing: true to release both managed and unmanaged resources; false to release only unmanaged resources.
  """
  pass
 def OnDataGridViewChanged(self,*args):
  """
  OnDataGridViewChanged(self: DataGridViewBand)

   Called when the band is associated with a different System.Windows.Forms.DataGridView.
  """
  pass
 def RaiseCellClick(self,*args):
  """
  RaiseCellClick(self: DataGridViewElement,e: DataGridViewCellEventArgs)

   Raises the System.Windows.Forms.DataGridView.CellClick event.

  

   e: A System.Windows.Forms.DataGridViewCellEventArgs that contains the event data.
  """
  pass
 def RaiseCellContentClick(self,*args):
  """
  RaiseCellContentClick(self: DataGridViewElement,e: DataGridViewCellEventArgs)

   Raises the System.Windows.Forms.DataGridView.CellContentClick event.

  

   e: A System.Windows.Forms.DataGridViewCellEventArgs that contains the event data.
  """
  pass
 def RaiseCellContentDoubleClick(self,*args):
  """
  RaiseCellContentDoubleClick(self: DataGridViewElement,e: DataGridViewCellEventArgs)

   Raises the System.Windows.Forms.DataGridView.CellContentDoubleClick event.

  

   e: A System.Windows.Forms.DataGridViewCellEventArgs that contains the event data.
  """
  pass
 def RaiseCellValueChanged(self,*args):
  """
  RaiseCellValueChanged(self: DataGridViewElement,e: DataGridViewCellEventArgs)

   Raises the System.Windows.Forms.DataGridView.CellValueChanged event.

  

   e: A System.Windows.Forms.DataGridViewCellEventArgs that contains the event data.
  """
  pass
 def RaiseDataError(self,*args):
  """
  RaiseDataError(self: DataGridViewElement,e: DataGridViewDataErrorEventArgs)

   Raises the System.Windows.Forms.DataGridView.DataError event.

  

   e: A System.Windows.Forms.DataGridViewDataErrorEventArgs that contains the event data.
  """
  pass
 def RaiseMouseWheel(self,*args):
  """
  RaiseMouseWheel(self: DataGridViewElement,e: MouseEventArgs)

   Raises the System.Windows.Forms.Control.MouseWheel event.

  

   e: A System.Windows.Forms.MouseEventArgs that contains the event data.
  """
  pass
 def ToString(self):
  """
  ToString(self: DataGridViewButtonColumn) -> str

   Returns: A System.String that describes the column.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __str__(self,*args):
  pass
 CellTemplate=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the template used to create new cells.



Get: CellTemplate(self: DataGridViewButtonColumn) -> DataGridViewCell



Set: CellTemplate(self: DataGridViewButtonColumn)=value

"""

 DefaultCellStyle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the column's default cell style.



Get: DefaultCellStyle(self: DataGridViewButtonColumn) -> DataGridViewCellStyle



Set: DefaultCellStyle(self: DataGridViewButtonColumn)=value

"""

 FlatStyle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the flat-style appearance of the button cells in the column.



Get: FlatStyle(self: DataGridViewButtonColumn) -> FlatStyle



Set: FlatStyle(self: DataGridViewButtonColumn)=value

"""

 HeaderCellCore=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the header cell of the System.Windows.Forms.DataGridViewBand.



"""

 IsRow=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the band represents a row.



"""

 Text=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the default text displayed on the button cell.



Get: Text(self: DataGridViewButtonColumn) -> str



Set: Text(self: DataGridViewButtonColumn)=value

"""

 UseColumnTextForButtonValue=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether the System.Windows.Forms.DataGridViewButtonColumn.Text property value is displayed as the button text for cells in this column.



Get: UseColumnTextForButtonValue(self: DataGridViewButtonColumn) -> bool



Set: UseColumnTextForButtonValue(self: DataGridViewButtonColumn)=value

"""





class ProgressChangedEventArgs(EventArgs):
 """
 Provides data for the System.ComponentModel.BackgroundWorker.ProgressChanged event.

 

 ProgressChangedEventArgs(progressPercentage: int,userState: object)
 """
 @staticmethod
 def __new__(self,progressPercentage,userState):
  """ __new__(cls: type,progressPercentage: int,userState: object) """
  pass
 ProgressPercentage=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the asynchronous task progress percentage.



Get: ProgressPercentage(self: ProgressChangedEventArgs) -> int



"""

 UserState=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a unique user state.



Get: UserState(self: ProgressChangedEventArgs) -> object



"""





class LayoutEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.Control.Layout event. This class cannot be inherited.

 

 LayoutEventArgs(affectedComponent: IComponent,affectedProperty: str)

 LayoutEventArgs(affectedControl: Control,affectedProperty: str)
 """
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type,affectedComponent: IComponent,affectedProperty: str)

  __new__(cls: type,affectedControl: Control,affectedProperty: str)
  """
  pass
 AffectedComponent=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.ComponentModel.Component affected by the layout change.



Get: AffectedComponent(self: LayoutEventArgs) -> IComponent



"""

 AffectedControl=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the child control affected by the change.



Get: AffectedControl(self: LayoutEventArgs) -> Control



"""

 AffectedProperty=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the property affected by the change.



Get: AffectedProperty(self: LayoutEventArgs) -> str



"""





class SendKeys(object):
 """ Provides methods for sending keystrokes to an application. """
 @staticmethod
 def Flush():
  """
  Flush()

   Processes all the Windows messages currently in the message queue.
  """
  pass
 @staticmethod
 def Send(keys):
  """
  Send(keys: str)

   Sends keystrokes to the active application.

  

   keys: The string of keystrokes to send.
  """
  pass
 @staticmethod
 def SendWait(keys):
  """
  SendWait(keys: str)

   Sends the given keys to the active application,and then waits for the messages to be processed.

  

   keys: The string of keystrokes to send.
  """
  pass



class ControlFlags(Enum,IComparable,IFormattable,IConvertible):
 """
 These flags affect the security descriptor behavior.

 

 enum (flags) ControlFlags,values: DiscretionaryAclAutoInherited (1024),DiscretionaryAclAutoInheritRequired (256),DiscretionaryAclDefaulted (8),DiscretionaryAclPresent (4),DiscretionaryAclProtected (4096),DiscretionaryAclUntrusted (64),GroupDefaulted (2),None (0),OwnerDefaulted (1),RMControlValid (16384),SelfRelative (32768),ServerSecurity (128),SystemAclAutoInherited (2048),SystemAclAutoInheritRequired (512),SystemAclDefaulted (32),SystemAclPresent (16),SystemAclProtected (8192)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 DiscretionaryAclAutoInherited=None
 DiscretionaryAclAutoInheritRequired=None
 DiscretionaryAclDefaulted=None
 DiscretionaryAclPresent=None
 DiscretionaryAclProtected=None
 DiscretionaryAclUntrusted=None
 GroupDefaulted=None
 None=None
 OwnerDefaulted=None
 RMControlValid=None
 SelfRelative=None
 ServerSecurity=None
 SystemAclAutoInherited=None
 SystemAclAutoInheritRequired=None
 SystemAclDefaulted=None
 SystemAclPresent=None
 SystemAclProtected=None
 value__=None




""" Transform Wrappers """

import math
import rpw
from rpw import DB
from rpw.base import BaseObjectWrapper

class Transform(BaseObjectWrapper):
    """
    `DB.Transform` Wrapper

    Attributes:
        _revit_object (DB.XYZ): Wrapped ``DB.Transform``
    """

    _revit_object_class = DB.Transform

    @classmethod
    def rotate_vector(cls, vector, rotation, center=None, axis=None, radians=False):
        """ Rotate a Vector or point

        Usage:
        >>> from rpw import db
        >>> db.Transform.rotate_vector(SomeVector, 90.0)

        Args:
            vector (``point-like``): Point like element.
            rotation (``float``): Rotation in degrees.
            center (``point-like``, optional): Center of rotation [default: 0,0,0]
            axis (``point-like``, optional): Axis of rotation [default: 0,0,1]
            radians (``bool``, optional): True for rotation angle is in radians [default: False]

        Returns:
            ``point-like``: Rotate Vector
            """
        XYZ = rpw.db.XYZ
        vector = XYZ(vector)
        if radians:
            angle_rad = rotation
        else:
            angle_rad = math.radians(rotation)
        axis =  XYZ(DB.XYZ(0,0,1)) if not axis else XYZ(axis)
        center =  XYZ(DB.XYZ(0,0,0)) if not center else XYZ(center)
        transform = cls._revit_object_class.CreateRotationAtPoint(
                                                        axis.unwrap(),
                                                        angle_rad,
                                                        center.unwrap())

        return XYZ(transform.OfVector(vector.unwrap()))

    @classmethod
    def move(cls, vector, object):
        """ Rotate a Vector by Degrees """

        raise NotImplemented
        # vector = XYZ(vector)
        # transform = cls._revit_object_class.CreateTranslation(vector)



class ListBindingConverter(TypeConverter):
 """
 Provides a type converter to convert System.Windows.Forms.Binding objects to and from various other representations.

 

 ListBindingConverter()
 """
 def CanConvertTo(self,*__args):
  """
  CanConvertTo(self: ListBindingConverter,context: ITypeDescriptorContext,destinationType: Type) -> bool

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   destinationType: A System.Type that represents the type you want to convert to.

   Returns: true if this converter can perform the conversion; otherwise,false.
  """
  pass
 def ConvertTo(self,*__args):
  """
  ConvertTo(self: ListBindingConverter,context: ITypeDescriptorContext,culture: CultureInfo,value: object,destinationType: Type) -> object

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   culture: A System.Globalization.CultureInfo. If null is passed,the current culture is assumed.

   value: The System.Object to convert.

   destinationType: The System.Type to convert the value parameter to.

   Returns: An System.Object that represents the converted value.
  """
  pass
 def CreateInstance(self,*__args):
  """
  CreateInstance(self: ListBindingConverter,context: ITypeDescriptorContext,propertyValues: IDictionary) -> object

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   propertyValues: An System.Collections.IDictionary of new property values.

   Returns: An System.Object representing the given System.Collections.IDictionary,or null if the object 

    cannot be created. This method always returns null.
  """
  pass
 def GetCreateInstanceSupported(self,context=None):
  """
  GetCreateInstanceSupported(self: ListBindingConverter,context: ITypeDescriptorContext) -> bool

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   Returns: true if changing a property on this object requires a call to 

    System.ComponentModel.TypeConverter.CreateInstance(System.Collections.IDictionary) to create a 

    new value; otherwise,false.
  """
  pass



class Pen(MarshalByRefObject,ISystemColorTracker,ICloneable,IDisposable):
 """
 Defines an object used to draw lines and curves. This class cannot be inherited.

 

 Pen(color: Color)

 Pen(color: Color,width: Single)

 Pen(brush: Brush)

 Pen(brush: Brush,width: Single)
 """
 def Clone(self):
  """
  Clone(self: Pen) -> object

  

   Creates an exact copy of this System.Drawing.Pen.

   Returns: An System.Object that can be cast to a System.Drawing.Pen.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: Pen)

   Releases all resources used by this System.Drawing.Pen.
  """
  pass
 def MemberwiseClone(self,*args):
  """
  MemberwiseClone(self: MarshalByRefObject,cloneIdentity: bool) -> MarshalByRefObject

  

   Creates a shallow copy of the current System.MarshalByRefObject object.

  

   cloneIdentity: false to delete the current System.MarshalByRefObject object's identity,which will cause the 

    object to be assigned a new identity when it is marshaled across a remoting boundary. A value of 

    false is usually appropriate. true to copy the current System.MarshalByRefObject object's 

    identity to its clone,which will cause remoting client calls to be routed to the remote server 

    object.

  

   Returns: A shallow copy of the current System.MarshalByRefObject object.

  MemberwiseClone(self: object) -> object

  

   Creates a shallow copy of the current System.Object.

   Returns: A shallow copy of the current System.Object.
  """
  pass
 def MultiplyTransform(self,matrix,order=None):
  """
  MultiplyTransform(self: Pen,matrix: Matrix,order: MatrixOrder)

   Multiplies the transformation matrix for this System.Drawing.Pen by the specified 

    System.Drawing.Drawing2D.Matrix in the specified order.

  

  

   matrix: The System.Drawing.Drawing2D.Matrix by which to multiply the transformation matrix.

   order: The order in which to perform the multiplication operation.

  MultiplyTransform(self: Pen,matrix: Matrix)

   Multiplies the transformation matrix for this System.Drawing.Pen by the specified 

    System.Drawing.Drawing2D.Matrix.

  

  

   matrix: The System.Drawing.Drawing2D.Matrix object by which to multiply the transformation matrix.
  """
  pass
 def ResetTransform(self):
  """
  ResetTransform(self: Pen)

   Resets the geometric transformation matrix for this System.Drawing.Pen to identity.
  """
  pass
 def RotateTransform(self,angle,order=None):
  """
  RotateTransform(self: Pen,angle: Single,order: MatrixOrder)

   Rotates the local geometric transformation by the specified angle in the specified order.

  

   angle: The angle of rotation.

   order: A System.Drawing.Drawing2D.MatrixOrder that specifies whether to append or prepend the rotation 

    matrix.

  

  RotateTransform(self: Pen,angle: Single)

   Rotates the local geometric transformation by the specified angle. This method prepends the 

    rotation to the transformation.

  

  

   angle: The angle of rotation.
  """
  pass
 def ScaleTransform(self,sx,sy,order=None):
  """
  ScaleTransform(self: Pen,sx: Single,sy: Single,order: MatrixOrder)

   Scales the local geometric transformation by the specified factors in the specified order.

  

   sx: The factor by which to scale the transformation in the x-axis direction.

   sy: The factor by which to scale the transformation in the y-axis direction.

   order: A System.Drawing.Drawing2D.MatrixOrder that specifies whether to append or prepend the scaling 

    matrix.

  

  ScaleTransform(self: Pen,sx: Single,sy: Single)

   Scales the local geometric transformation by the specified factors. This method prepends the 

    scaling matrix to the transformation.

  

  

   sx: The factor by which to scale the transformation in the x-axis direction.

   sy: The factor by which to scale the transformation in the y-axis direction.
  """
  pass
 def SetLineCap(self,startCap,endCap,dashCap):
  """
  SetLineCap(self: Pen,startCap: LineCap,endCap: LineCap,dashCap: DashCap)

   Sets the values that determine the style of cap used to end lines drawn by this 

    System.Drawing.Pen.

  

  

   startCap: A System.Drawing.Drawing2D.LineCap that represents the cap style to use at the beginning of 

    lines drawn with this System.Drawing.Pen.

  

   endCap: A System.Drawing.Drawing2D.LineCap that represents the cap style to use at the end of lines 

    drawn with this System.Drawing.Pen.

  

   dashCap: A System.Drawing.Drawing2D.LineCap that represents the cap style to use at the beginning or end 

    of dashed lines drawn with this System.Drawing.Pen.
  """
  pass
 def TranslateTransform(self,dx,dy,order=None):
  """
  TranslateTransform(self: Pen,dx: Single,dy: Single,order: MatrixOrder)

   Translates the local geometric transformation by the specified dimensions in the specified order.

  

   dx: The value of the translation in x.

   dy: The value of the translation in y.

   order: The order (prepend or append) in which to apply the translation.

  TranslateTransform(self: Pen,dx: Single,dy: Single)

   Translates the local geometric transformation by the specified dimensions. This method prepends 

    the translation to the transformation.

  

  

   dx: The value of the translation in x.

   dy: The value of the translation in y.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type,color: Color)

  __new__(cls: type,color: Color,width: Single)

  __new__(cls: type,brush: Brush)

  __new__(cls: type,brush: Brush,width: Single)
  """
  pass
 Alignment=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the alignment for this System.Drawing.Pen.



Get: Alignment(self: Pen) -> PenAlignment



Set: Alignment(self: Pen)=value

"""

 Brush=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the System.Drawing.Brush that determines attributes of this System.Drawing.Pen.



Get: Brush(self: Pen) -> Brush



Set: Brush(self: Pen)=value

"""

 Color=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the color of this System.Drawing.Pen.



Get: Color(self: Pen) -> Color



Set: Color(self: Pen)=value

"""

 CompoundArray=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an array of values that specifies a compound pen. A compound pen draws a compound line made up of parallel lines and spaces.



Get: CompoundArray(self: Pen) -> Array[Single]



Set: CompoundArray(self: Pen)=value

"""

 CustomEndCap=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a custom cap to use at the end of lines drawn with this System.Drawing.Pen.



Get: CustomEndCap(self: Pen) -> CustomLineCap



Set: CustomEndCap(self: Pen)=value

"""

 CustomStartCap=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a custom cap to use at the beginning of lines drawn with this System.Drawing.Pen.



Get: CustomStartCap(self: Pen) -> CustomLineCap



Set: CustomStartCap(self: Pen)=value

"""

 DashCap=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the cap style used at the end of the dashes that make up dashed lines drawn with this System.Drawing.Pen.



Get: DashCap(self: Pen) -> DashCap



Set: DashCap(self: Pen)=value

"""

 DashOffset=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the distance from the start of a line to the beginning of a dash pattern.



Get: DashOffset(self: Pen) -> Single



Set: DashOffset(self: Pen)=value

"""

 DashPattern=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets an array of custom dashes and spaces.



Get: DashPattern(self: Pen) -> Array[Single]



Set: DashPattern(self: Pen)=value

"""

 DashStyle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the style used for dashed lines drawn with this System.Drawing.Pen.



Get: DashStyle(self: Pen) -> DashStyle



Set: DashStyle(self: Pen)=value

"""

 EndCap=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the cap style used at the end of lines drawn with this System.Drawing.Pen.



Get: EndCap(self: Pen) -> LineCap



Set: EndCap(self: Pen)=value

"""

 LineJoin=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the join style for the ends of two consecutive lines drawn with this System.Drawing.Pen.



Get: LineJoin(self: Pen) -> LineJoin



Set: LineJoin(self: Pen)=value

"""

 MiterLimit=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the limit of the thickness of the join on a mitered corner.



Get: MiterLimit(self: Pen) -> Single



Set: MiterLimit(self: Pen)=value

"""

 PenType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the style of lines drawn with this System.Drawing.Pen.



Get: PenType(self: Pen) -> PenType



"""

 StartCap=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the cap style used at the beginning of lines drawn with this System.Drawing.Pen.



Get: StartCap(self: Pen) -> LineCap



Set: StartCap(self: Pen)=value

"""

 Transform=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a copy of the geometric transformation for this System.Drawing.Pen.



Get: Transform(self: Pen) -> Matrix



Set: Transform(self: Pen)=value

"""

 Width=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the width of this System.Drawing.Pen,in units of the System.Drawing.Graphics object used for drawing.



Get: Width(self: Pen) -> Single



Set: Width(self: Pen)=value

"""





class FileMode(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies how the operating system should open a file.

 

 enum FileMode,values: Append (6),Create (2),CreateNew (1),Open (3),OpenOrCreate (4),Truncate (5)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Append=None
 Create=None
 CreateNew=None
 Open=None
 OpenOrCreate=None
 Truncate=None
 value__=None

